__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

"""
The SSDP/UPNP implementations has been largely inspired by https://gist.github.com/schlamar/2428250

For a spec see http://www.upnp.org/specs/arch/UPnP-arch-DeviceArchitecture-v1.0.pdf
"""

import collections
import platform
import socket
import time

import flask
import zeroconf
from flask_babel import gettext

import octoprint.plugin
import octoprint.util


def __plugin_load__():
    plugin = DiscoveryPlugin()

    global __plugin_implementation__
    __plugin_implementation__ = plugin

    global __plugin_helpers__
    __plugin_helpers__ = {
        "ssdp_browse": plugin.ssdp_browse,
        "zeroconf_browse": plugin.zeroconf_browse,
        "zeroconf_register": plugin.zeroconf_register,
        "zeroconf_unregister": plugin.zeroconf_unregister,
    }


class DiscoveryPlugin(
    octoprint.plugin.StartupPlugin,
    octoprint.plugin.ShutdownPlugin,
    octoprint.plugin.BlueprintPlugin,
    octoprint.plugin.SettingsPlugin,
):

    ssdp_multicast_addr = "239.255.255.250"  # IPv6: ff0X::c

    ssdp_multicast_port = 1900

    ssdp_server = "{}/{} UPnP/1.0 OctoPrint/{}".format(
        platform.system(), platform.version(), octoprint.__version__
    )

    # noinspection PyMissingConstructor
    def __init__(self):
        self.host = None
        self.port = None

        # zeroconf
        self._zeroconf = None
        self._zeroconf_registrations = collections.defaultdict(list)

        # upnp/ssdp
        self._ssdp_monitor_active = False
        self._ssdp_monitor_thread = None
        self._ssdp_notify_timeout = 30
        self._ssdp_last_notify = 0

    def initialize(self):
        self._zeroconf = zeroconf.Zeroconf(interfaces=self.get_interface_addresses())

    ##~~ SettingsPlugin API

    def get_settings_defaults(self):
        return {
            "publicHost": None,
            "publicPort": None,
            "pathPrefix": None,
            "httpUsername": None,
            "httpPassword": None,
            "upnpUuid": None,
            "zeroConf": [],
            "model": {
                "name": None,
                "description": None,
                "number": None,
                "url": None,
                "serial": None,
                "vendor": None,
                "vendorUrl": None,
            },
            "addresses": None,
            "ignoredAddresses": None,
            "interfaces": None,
            "ignoredInterfaces": None,
        }

    ##~~ BlueprintPlugin API -- used for providing the SSDP device descriptor XML

    @octoprint.plugin.BlueprintPlugin.route("/discovery.xml", methods=["GET"])
    def discovery(self):
        self._logger.debug("Rendering discovery.xml")

        modelName = self._settings.get(["model", "name"])
        if not modelName:
            import octoprint.server

            modelName = octoprint.server.DISPLAY_VERSION

        vendor = self._settings.get(["model", "vendor"])
        vendorUrl = self._settings.get(["model", "vendorUrl"])
        if not vendor:
            vendor = "The OctoPrint Project"
            vendorUrl = "http://www.octoprint.org/"

        response = flask.make_response(
            flask.render_template(
                "discovery.xml.jinja2",
                friendlyName=self.get_instance_name(),
                manufacturer=vendor,
                manufacturerUrl=vendorUrl,
                modelName=modelName,
                modelDescription=self._settings.get(["model", "description"]),
                modelNumber=self._settings.get(["model", "number"]),
                modelUrl=self._settings.get(["model", "url"]),
                serialNumber=self._settings.get(["model", "serial"]),
                uuid=self.get_uuid(),
                presentationUrl=flask.url_for("index", _external=True),
            )
        )
        response.headers["Content-Type"] = "application/xml"
        return response

    def is_blueprint_protected(self):
        return False

    def is_blueprint_csrf_protected(self):
        return True

    ##~~ StartupPlugin API -- used for registering OctoPrint's Zeroconf and SSDP services upon application startup

    def on_startup(self, host, port):
        public_host = self._settings.get(["publicHost"])
        if public_host:
            host = public_host
        public_port = self._settings.get(["publicPort"])
        if public_port:
            port = public_port

        self.host = host
        self.port = port

        # Zeroconf
        instance_name = self.get_instance_name()
        self.zeroconf_register(
            "_http._tcp", instance_name, txt_record=self._create_http_txt_record_dict()
        )
        self.zeroconf_register(
            "_octoprint._tcp",
            instance_name,
            txt_record=self._create_octoprint_txt_record_dict(),
        )
        for zc in self._settings.get(["zeroConf"]):
            if "service" in zc:
                self.zeroconf_register(
                    zc["service"],
                    zc.get("name", instance_name),
                    port=zc.get("port"),
                    txt_record=zc.get("txtRecord"),
                )

        # SSDP
        self._ssdp_register()

    ##~~ ShutdownPlugin API -- used for unregistering OctoPrint's Zeroconf and SSDP service upon application shutdown

    def on_shutdown(self):
        registrations = list(self._zeroconf_registrations.keys())
        for key in registrations:
            reg_type, port = key
            self.zeroconf_unregister(reg_type, port)

        self._ssdp_unregister()

    ##~~ helpers

    # ZeroConf

    def _format_zeroconf_service_type(self, service_type):
        if not service_type.endswith("."):
            service_type += "."
        if not service_type.endswith("local."):
            service_type += "local."
        return service_type

    def _format_zeroconf_name(self, name, service_type):
        service_type = self._format_zeroconf_service_type(service_type)
        return f"{name}.{service_type}"

    def _format_zeroconf_txt(self, record):
        result = {}
        if not record:
            return result

        for key, value in record.items():
            result[octoprint.util.to_bytes(key)] = octoprint.util.to_bytes(value)
        return result

    def zeroconf_register(self, reg_type, name=None, port=None, txt_record=None):
        """
        Registers a new service with Zeroconf/Bonjour/Avahi.

        :param reg_type: type of service to register, e.g. "_gntp._tcp"
        :param name: displayable name of the service, if not given defaults to the OctoPrint instance name
        :param port: port to register for the service, if not given defaults to OctoPrint's (public) port
        :param txt_record: optional txt record to attach to the service, dictionary of key-value-pairs
        """

        if not name:
            name = self.get_instance_name()
        if not port:
            port = self.port

        reg_type = self._format_zeroconf_service_type(reg_type)
        name = self._format_zeroconf_name(name, reg_type)
        txt_record = self._format_zeroconf_txt(txt_record)

        key = (reg_type, port)
        addresses = list(
            map(lambda x: socket.inet_aton(x), self.get_interface_addresses())
        )

        try:
            info = zeroconf.ServiceInfo(
                reg_type,
                name,
                addresses=addresses,
                port=port,
                server=f"{socket.gethostname()}.local.",
                properties=txt_record,
            )
            self._zeroconf.register_service(info, allow_name_change=True)
            self._zeroconf_registrations[key].append(info)
            self._logger.info(f"Registered '{name}' for {reg_type}")
        except Exception:
            self._logger.exception(
                f"Could not register {name} for {reg_type} on port {port}"
            )

    def zeroconf_unregister(self, reg_type, port=None):
        """
        Unregisters a previously registered Zeroconf/Bonjour/Avahi service identified by service and port.

        :param reg_type: the type of the service to be unregistered
        :param port: the port of the service to be unregistered, defaults to OctoPrint's (public) port if not given
        :return:
        """

        if not port:
            port = self.port
        reg_type = self._format_zeroconf_service_type(reg_type)

        key = (reg_type, port)
        if key not in self._zeroconf_registrations:
            return

        infos = self._zeroconf_registrations.pop(key)
        try:
            for info in infos:
                self._zeroconf.unregister_service(info)
            self._logger.debug(f"Unregistered {reg_type} on port {port}")
        except Exception:
            self._logger.exception(
                f"Could not (fully) unregister {reg_type} on port {port}"
            )

    def zeroconf_browse(
        self, service_type, block=True, callback=None, browse_timeout=5, resolve_timeout=5
    ):
        """
        Browses for services on the local network providing the specified service type. Can be used either blocking or
        non-blocking.

        The non-blocking version (default behaviour) will not return until the lookup has completed and
        return all results that were found.

        For non-blocking version, set `block` to `False` and provide a `callback` to be called once the lookup completes.
        If no callback is provided in non-blocking mode, a ValueError will be raised.

        The results are provided as a list of discovered services, with each service being described by a dictionary
        with the following keys:

          * `name`: display name of the service
          * `host`: host name of the service
          * `post`: port the service is listening on
          * `txt_record`: TXT record of the service as a dictionary, exact contents depend on the service

        Callbacks will be called with that list as the single parameter supplied to them. Thus, the following is an
        example for a valid callback:

            def browse_callback(results):
              for result in results:
                print "Name: {name}, Host: {host}, Port: {port}, TXT: {txt_record!r}".format(**result)

        :param service_type: the service type to browse for
        :param block: whether to block, defaults to True
        :param callback: callback to call once lookup has completed, must be set when `block` is set to `False`
        :param browse_timeout: timeout for browsing operation
        :param resolve_timeout: timeout for resolving operations for discovered records
        :return: if `block` is `True` a list of the discovered services, an empty list otherwise (results will then be
                 supplied to the callback instead)
        """

        import threading

        if not block and not callback:
            raise ValueError("Non-blocking mode but no callback given")

        service_type = self._format_zeroconf_service_type(service_type)

        result = []
        result_available = threading.Event()
        result_available.clear()

        class ZeroconfListener:
            def __init__(self, logger):
                self._logger = logger

            def add_service(self, zeroconf, type, name):
                self._logger.debug(
                    "Got a browsing result for Zeroconf resolution of {}, resolving...".format(
                        type
                    )
                )
                info = zeroconf.get_service_info(
                    type, name, timeout=resolve_timeout * 1000
                )
                if info:

                    def to_result(info, address):
                        n = info.name[: -(len(type) + 1)]
                        p = info.port

                        self._logger.debug(
                            "Resolved a result for Zeroconf resolution of {}: {} @ {}:{}".format(
                                type, n, address, p
                            )
                        )

                        return {
                            "name": n,
                            "host": address,
                            "port": p,
                            "txt_record": info.properties,
                        }

                    for address in map(lambda x: socket.inet_ntoa(x), info.addresses):
                        result.append(to_result(info, address))

        self._logger.debug(f"Browsing Zeroconf for {service_type}")

        def browse():
            listener = ZeroconfListener(self._logger)
            browser = zeroconf.ServiceBrowser(self._zeroconf, service_type, listener)
            time.sleep(browse_timeout)
            browser.cancel()

            if callback:
                callback(result)
            result_available.set()

        browse_thread = threading.Thread(target=browse)
        browse_thread.daemon = True
        browse_thread.start()

        if block:
            result_available.wait()
            return result
        else:
            return []

    # SSDP/UPNP

    def ssdp_browse(self, query, block=True, callback=None, timeout=1, retries=5):
        """
        Browses for UPNP services matching the supplied query. Can be used either blocking or
        non-blocking.

        The non-blocking version (default behaviour) will not return until the lookup has completed and
        return all results that were found.

        For non-blocking version, set `block` to `False` and provide a `callback` to be called once the lookup completes.
        If no callback is provided in non-blocking mode, a ValueError will be raised.

        The results are provided as a list of discovered locations of device descriptor files.

        Callbacks will be called with that list as the single parameter supplied to them. Thus, the following is an
        example for a valid callback:

            def browse_callback(results):
              for result in results:
                print("Location: {}".format(result))

        :param query: the SSDP query to send, e.g. "upnp:rootdevice" to search for all devices
        :param block: whether to block, defaults to True
        :param callback: callback to call in non-blocking mode when lookup has finished, must be set if block is False
        :param timeout: timeout in seconds to wait for replies to the M-SEARCH query per interface, defaults to 1
        :param retries: number of retries to perform the lookup on all interfaces, defaults to 5
        :return: if `block` is `True` a list of the discovered devices, an empty list otherwise (results will then be
                 supplied to the callback instead)
        """

        import io
        import threading
        from http.client import HTTPResponse

        class Response(HTTPResponse):
            def __init__(self, response_text):
                self.fp = io.BytesIO(response_text)
                self.debuglevel = 0
                self.strict = 0
                self.msg = None
                self._method = None
                self.begin()

        result = []
        result_available = threading.Event()
        result_available.clear()

        def browse():
            socket.setdefaulttimeout(timeout)

            search_message = "".join(
                [
                    "M-SEARCH * HTTP/1.1\r\n",
                    "ST: {query}\r\n",
                    "MX: 3\r\n",
                    'MAN: "ssdp:discovery"\r\n',
                    "HOST: {mcast_addr}:{mcast_port}\r\n\r\n",
                ]
            )

            for _ in range(retries):
                for addr in self.get_interface_addresses():
                    try:
                        sock = socket.socket(
                            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
                        )
                        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
                        sock.bind((addr, 0))

                        message = search_message.format(
                            query=query,
                            mcast_addr=self.__class__.ssdp_multicast_addr,
                            mcast_port=self.__class__.ssdp_multicast_port,
                        )
                        for _ in range(2):
                            sock.sendto(
                                message.encode("utf-8"),
                                (
                                    self.__class__.ssdp_multicast_addr,
                                    self.__class__.ssdp_multicast_port,
                                ),
                            )

                        try:
                            data = sock.recv(1024)
                        except socket.timeout:
                            pass
                        else:
                            response = Response(data)

                            result.append(response.getheader("Location"))
                    except Exception:
                        pass

            if callback:
                callback(result)
            result_available.set()

        browse_thread = threading.Thread(target=browse)
        browse_thread.daemon = True
        browse_thread.start()

        if block:
            result_available.wait()
            return result
        else:
            return []

    ##~~ internals

    # Zeroconf

    def _create_http_txt_record_dict(self):
        """
        Creates a TXT record for the _http._tcp Zeroconf service supplied by this OctoPrint instance.

        Defines the keys for _http._tcp as defined in http://www.dns-sd.org/txtrecords.html

        :return: a dictionary containing the defined key-value-pairs, ready to be turned into a TXT record
        """

        # determine path entry
        path = "/"
        if self._settings.get(["pathPrefix"]):
            path = self._settings.get(["pathPrefix"])
        else:
            prefix = self._settings.global_get(
                ["server", "reverseProxy", "prefixFallback"]
            )
            if prefix:
                path = prefix

        # fetch username and password (if set)
        username = self._settings.get(["httpUsername"])
        password = self._settings.get(["httpPassword"])

        entries = {"path": path}

        if username and password:
            entries.update({"u": username, "p": password})

        return entries

    def _create_octoprint_txt_record_dict(self):
        """
        Creates a TXT record for the _octoprint._tcp Zeroconf service supplied by this OctoPrint instance.

        The following keys are defined:

          * `path`: path prefix to actual OctoPrint instance, inherited from _http._tcp
          * `u`: username if HTTP Basic Auth is used, optional, inherited from _http._tcp
          * `p`: password if HTTP Basic Auth is used, optional, inherited from _http._tcp
          * `version`: OctoPrint software version
          * `api`: OctoPrint API version
          * `model`: Model of the device that is running OctoPrint
          * `vendor`: Vendor of the device that is running OctoPrint

        :return: a dictionary containing the defined key-value-pairs, ready to be turned into a TXT record
        """
        import octoprint.server
        import octoprint.server.api

        entries = self._create_http_txt_record_dict()
        entries.update(
            version=octoprint.server.VERSION,
            api=octoprint.server.api.VERSION,
            uuid=self.get_uuid(),
        )

        modelName = self._settings.get(["model", "name"])
        if modelName:
            entries.update(model=modelName)
        vendor = self._settings.get(["model", "vendor"])
        if vendor:
            entries.update(vendor=vendor)

        return entries

    # SSDP/UPNP

    def _ssdp_register(self):
        """
        Registers the OctoPrint instance as basic service with a presentation URL pointing to the web interface
        """

        import threading

        self._ssdp_monitor_active = True

        self._ssdp_monitor_thread = threading.Thread(
            target=self._ssdp_monitor, kwargs={"timeout": self._ssdp_notify_timeout}
        )
        self._ssdp_monitor_thread.daemon = True
        self._ssdp_monitor_thread.start()

    def _ssdp_unregister(self):
        """
        Unregisters the OctoPrint instance again
        """

        self._ssdp_monitor_active = False
        if self.host and self.port:
            for _ in range(2):
                self._ssdp_notify(alive=False)

    def _ssdp_notify(self, alive=True):
        """
        Sends an SSDP notify message across the connected networks.

        :param alive: True to send an "ssdp:alive" message, False to send an "ssdp:byebye" message
        """

        if (
            alive
            and self._ssdp_last_notify + self._ssdp_notify_timeout > time.monotonic()
        ):
            # we just sent an alive, no need to send another one now
            return

        if alive and not self._ssdp_monitor_active:
            # the monitor already shut down, alive messages don't make sense anymore as byebye will shortly follow
            return

        for addr in self.get_interface_addresses():
            try:
                sock = socket.socket(
                    socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
                )
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
                sock.bind((addr, 0))

                location = "http://{addr}:{port}/plugin/discovery/discovery.xml".format(
                    addr=addr, port=self.port
                )

                self._logger.debug(
                    "Sending NOTIFY {} via {}".format(
                        "alive" if alive else "byebye", addr
                    )
                )
                notify_message = "".join(
                    [
                        "NOTIFY * HTTP/1.1\r\n",
                        "Server: {server}\r\n",
                        "Cache-Control: max-age=900\r\n",
                        "Location: {location}\r\n",
                        "NTS: {nts}\r\n",
                        "NT: {nt}\r\n",
                        "USN: {usn}\r\n",
                        "HOST: {mcast_addr}:{mcast_port}\r\n\r\n",
                    ]
                )
                uuid = self.get_uuid()
                for nt, usn in (
                    ("upnp:rootdevice", "uuid:{uuid}::upnp:rootdevice"),
                    ("uuid:{}", "uuid:{uuid}"),
                    (
                        "urn:schemas-upnp-org:device:basic:1",
                        "uuid:{uuid}::url:schemas-upnp-org:device:basic:1",
                    ),
                ):
                    message = notify_message.format(
                        uuid=uuid,
                        location=location,
                        nts="ssdp:alive" if alive else "ssdp:byebye",
                        nt=nt.format(uuid=uuid),
                        usn=usn.format(uuid=uuid),
                        server=self.ssdp_server,
                        mcast_addr=self.ssdp_multicast_addr,
                        mcast_port=self.ssdp_multicast_port,
                    )
                    for _ in range(2):
                        # send twice, stuff might get lost, it's only UDP
                        sock.sendto(
                            message.encode("utf-8"),
                            (self.ssdp_multicast_addr, self.ssdp_multicast_port),
                        )
            except Exception:
                pass

        self._ssdp_last_notify = time.monotonic()

    def _ssdp_monitor(self, timeout=5):
        """
        Monitor thread that listens on the multicast address for M-SEARCH requests and answers them if they are relevant

        :param timeout: timeout after which to stop waiting for M-SEARCHs for a short while in order to put out an
                        alive message
        """

        from http.server import BaseHTTPRequestHandler
        from io import BytesIO

        socket.setdefaulttimeout(timeout)

        location_message = "".join(
            [
                "HTTP/1.1 200 OK\r\n",
                "ST: upnp:rootdevice\r\n",
                "USN: uuid:{uuid}::upnp:rootdevice\r\n",
                "Location: {location}\r\n",
                "Cache-Control: max-age=60\r\n\r\n",
            ]
        )

        class Request(BaseHTTPRequestHandler):

            # noinspection PyMissingConstructor
            def __init__(self, request_text):
                self.rfile = BytesIO(request_text)
                self.raw_requestline = self.rfile.readline()
                self.error_code = self.error_message = None
                self.parse_request()

            def send_error(self, code, message=None):
                self.error_code = code
                self.error_message = message

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        sock.bind(("", self.__class__.ssdp_multicast_port))

        sock.setsockopt(
            socket.IPPROTO_IP,
            socket.IP_ADD_MEMBERSHIP,
            socket.inet_aton(self.__class__.ssdp_multicast_addr)
            + socket.inet_aton("0.0.0.0"),
        )

        self._logger.info(f"Registered {self.get_instance_name()} for SSDP")

        self._ssdp_notify(alive=True)

        try:
            while self._ssdp_monitor_active:
                try:
                    data, address = sock.recvfrom(4096)
                    request = Request(data)
                    if (
                        not request.error_code
                        and request.command == "M-SEARCH"
                        and request.path == "*"
                        and (
                            request.headers["ST"] == "upnp:rootdevice"
                            or request.headers["ST"] == "ssdp:all"
                        )
                        and request.headers["MAN"] == '"ssdp:discover"'
                    ):
                        interface_address = octoprint.util.address_for_client(
                            *address,
                            addresses=self._settings.get(["addresses"]),
                            interfaces=self._settings.get(["interfaces"]),
                        )
                        if not interface_address:
                            continue

                        message = location_message.format(
                            uuid=self.get_uuid(),
                            location="http://{host}:{port}/plugin/discovery/discovery.xml".format(
                                host=interface_address, port=self.port
                            ),
                        )
                        sock.sendto(message.encode("utf-8"), address)
                        self._logger.debug(
                            "Sent M-SEARCH reply for {path} and {st} to {address!r}".format(
                                path=request.path,
                                st=request.headers["ST"],
                                address=address,
                            )
                        )
                except socket.timeout:
                    pass
                finally:
                    self._ssdp_notify(alive=True)
        finally:
            try:
                sock.close()
            except Exception:
                pass

    ##~~ helpers

    def get_uuid(self):
        upnpUuid = self._settings.get(["upnpUuid"])
        if upnpUuid is None:
            import uuid

            upnpUuid = str(uuid.uuid4())
            self._settings.set(["upnpUuid"], upnpUuid)
            self._settings.save()
        return upnpUuid

    def get_instance_name(self):
        name = self._settings.global_get(["appearance", "name"])
        if name:
            return f'OctoPrint instance "{name}"'
        else:
            return f"OctoPrint instance on {socket.gethostname()}"

    def get_interface_addresses(self):
        addresses = self._settings.get(["addresses"])
        if addresses:
            return addresses
        else:
            return list(
                octoprint.util.interface_addresses(
                    interfaces=self._settings.get(["interfaces"]),
                    ignored=self._settings.get(["ignoredInterfaces"]),
                )
            )


__plugin_name__ = "Discovery"
__plugin_author__ = "Gina Häußge"
__plugin_url__ = "https://docs.octoprint.org/en/master/bundledplugins/discovery.html"
__plugin_description__ = (
    "Makes the OctoPrint instance discoverable via Bonjour/Avahi/Zeroconf and uPnP"
)
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin your OctoPrint instance will no longer be "
    "discoverable on the network via Bonjour and uPnP."
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=3.7,<4"



"""
This module bundles all of OctoPrint's supported plugin implementation types as well as their common parent
class, :class:`OctoPrintPlugin`.

Please note that the plugin implementation types are documented in the section
:ref:`Available plugin mixins <sec-plugins-mixins>`.

.. autoclass:: OctoPrintPlugin
   :show-inheritance:
   :members:

.. autoclass:: ReloadNeedingPlugin
   :show-inheritance:
   :members:

"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


from .core import Plugin, RestartNeedingPlugin, SortablePlugin


class OctoPrintPlugin(Plugin):
    """
    The parent class of all OctoPrint plugin mixins.

    .. attribute:: _plugin_manager

       The :class:`~octoprint.plugin.core.PluginManager` instance. Injected by the plugin core system upon
       initialization of the implementation.

    .. attribute:: _printer_profile_manager

       The :class:`~octoprint.printer.profile.PrinterProfileManager` instance. Injected by the plugin core system upon
       initialization of the implementation.

    .. attribute:: _event_bus

       The :class:`~octoprint.events.EventManager` instance. Injected by the plugin core system upon initialization of
       the implementation.

    .. attribute:: _analysis_queue

       The :class:`~octoprint.filemanager.analysis.AnalysisQueue` instance. Injected by the plugin core system upon
       initialization of the implementation.

    .. attribute:: _slicing_manager

       The :class:`~octoprint.slicing.SlicingManager` instance. Injected by the plugin core system upon initialization
       of the implementation.

    .. attribute:: _file_manager

       The :class:`~octoprint.filemanager.FileManager` instance. Injected by the plugin core system upon initialization
       of the implementation.

    .. attribute:: _printer

       The :class:`~octoprint.printer.PrinterInterface` instance. Injected by the plugin core system upon initialization
       of the implementation.

    .. attribute:: _app_session_manager

       The :class:`~octoprint.access.users.SessionManager` instance. Injected by the plugin core system upon initialization of
       the implementation.

    .. attribute:: _plugin_lifecycle_manager

       The :class:`~octoprint.server.LifecycleManager` instance. Injected by the plugin core system upon initialization
       of the implementation.

    .. attribute:: _user_manager

       The :class:`~octoprint.access.users.UserManager` instance. Injected by the plugin core system upon initialization
       of the implementation.

    .. attribute:: _connectivity_checker

       The :class:`~octoprint.util.ConnectivityChecker` instance. Injected by the plugin core system upon initialization
       of the implementation.

    .. attribute:: _data_folder

       Path to the data folder for the plugin to use for any data it might have to persist. Should always be accessed
       through :meth:`get_plugin_data_folder` since that function will also ensure that the data folder actually exists
       and if not creating it before returning it. Injected by the plugin core system upon initialization of the
       implementation.
    """

    # noinspection PyMissingConstructor
    def __init__(self):
        self._plugin_manager = None
        self._printer_profile_manager = None
        self._event_bus = None
        self._analysis_queue = None
        self._slicing_manager = None
        self._file_manager = None
        self._printer = None
        self._app_session_manager = None
        self._plugin_lifecycle_manager = None
        self._user_manager = None
        self._connectivity_checker = None
        self._data_folder = None

    def get_plugin_data_folder(self):
        """
        Retrieves the path to a data folder specifically for the plugin, ensuring it exists and if not creating it
        before returning it.

        Plugins may use this folder for storing additional data they need for their operation.
        """
        if self._data_folder is None:
            raise RuntimeError(
                "self._plugin_data_folder is None, has the plugin been initialized yet?"
            )

        import os

        if not os.path.isdir(self._data_folder):
            os.makedirs(self._data_folder)
        return self._data_folder

    def on_plugin_pending_uninstall(self):
        """
        Called by the plugin manager when the plugin is pending uninstall. Override this to react to the event.

        NOT called during plugin uninstalls triggered outside of OctoPrint!
        """
        pass


class ReloadNeedingPlugin(Plugin):
    """
    Mixin for plugin types that need a reload of the UI after enabling/disabling them.
    """


class EnvironmentDetectionPlugin(OctoPrintPlugin, RestartNeedingPlugin):
    """
    .. versionadded:: 1.3.6
    """

    def get_additional_environment(self):
        pass

    def on_environment_detected(self, environment, *args, **kwargs):
        pass


class StartupPlugin(OctoPrintPlugin, SortablePlugin):
    """
    The ``StartupPlugin`` allows hooking into the startup of OctoPrint. It can be used to start up additional services
    on or just after the startup of the server.

    ``StartupPlugin`` is a :class:`~octoprint.plugin.core.SortablePlugin` and provides
    sorting contexts for :meth:`~octoprint.plugin.StartupPlugin.on_startup` as well as
    :meth:`~octoprint.plugin.StartupPlugin.on_after_startup`.
    """

    def on_startup(self, host, port):
        """
        Called just before the server is actually launched. Plugins get supplied with the ``host`` and ``port`` the server
        will listen on. Note that the ``host`` may be ``0.0.0.0`` if it will listen on all interfaces, so you can't just
        blindly use this for constructing publicly reachable URLs. Also note that when this method is called, the server
        is not actually up yet and none of your plugin's APIs or blueprints will be reachable yet. If you need to be
        externally reachable, use :func:`on_after_startup` instead or additionally.

        .. warning::

           Do not perform long-running or even blocking operations in your implementation or you **will** block and break the server.

        The relevant sorting context is ``StartupPlugin.on_startup``.

        :param string host: the host the server will listen on, may be ``0.0.0.0``
        :param int port:    the port the server will listen on
        """

        pass

    def on_after_startup(self):
        """
        Called just after launch of the server, so when the listen loop is actually running already.

        .. warning::

           Do not perform long-running or even blocking operations in your implementation or you **will** block and break the server.

        The relevant sorting context is ``StartupPlugin.on_after_startup``.
        """

        pass


class ShutdownPlugin(OctoPrintPlugin, SortablePlugin):
    """
    The ``ShutdownPlugin`` allows hooking into the shutdown of OctoPrint. It's usually used in conjunction with the
    :class:`StartupPlugin` mixin, to cleanly shut down additional services again that where started by the :class:`StartupPlugin`
    part of the plugin.

    ``ShutdownPlugin`` is a :class:`~octoprint.plugin.core.SortablePlugin` and provides a sorting context for
    :meth:`~octoprint.plugin.ShutdownPlugin.on_shutdown`.
    """

    def on_shutdown(self):
        """
        Called upon the imminent shutdown of OctoPrint.

        .. warning::

           Do not perform long-running or even blocking operations in your implementation or you **will** block and break the server.

        The relevant sorting context is ``ShutdownPlugin.on_shutdown``.
        """
        pass


class AssetPlugin(OctoPrintPlugin, RestartNeedingPlugin):
    """
    The ``AssetPlugin`` mixin allows plugins to define additional static assets such as JavaScript or CSS files to
    be automatically embedded into the pages delivered by the server to be used within the client sided part of
    the plugin.

    A typical usage of the ``AssetPlugin`` functionality is to embed a custom view model to be used by templates injected
    through a :class:`TemplatePlugin`.

    ``AssetPlugin`` is a :class:`~octoprint.plugins.core.RestartNeedingPlugin`.
    """

    def get_asset_folder(self):
        """
        Defines the folder where the plugin stores its static assets as defined in :func:`get_assets`. Override this if
        your plugin stores its assets at some other place than the ``static`` sub folder in the plugin base directory.

        :return string: the absolute path to the folder where the plugin stores its static assets
        """
        import os

        return os.path.join(self._basefolder, "static")

    def get_assets(self):
        """
        Defines the static assets the plugin offers. The following asset types are recognized and automatically
        imported at the appropriate places to be available:

        js
           JavaScript files, such as additional view models
        jsclient
           JavaScript files containing additional parts for the JS Client Library (since 1.3.10)
        css
           CSS files with additional styles, will be embedded into delivered pages when not running in LESS mode.
        less
           LESS files with additional styles, will be embedded into delivered pages when running in LESS mode.

        The expected format to be returned is a dictionary mapping one or more of these keys to a list of files of that
        type, the files being represented as relative paths from the asset folder as defined via :func:`get_asset_folder`.
        Example:

        .. code-block:: python

           def get_assets(self):
               return dict(
                   js=['js/my_file.js', 'js/my_other_file.js'],
                   clientjs=['clientjs/my_file.js'],
                   css=['css/my_styles.css'],
                   less=['less/my_styles.less']
                )

        The assets will be made available by OctoPrint under the URL ``/plugin/<plugin identifier>/static/<path>``, with
        ``plugin identifier`` being the plugin's identifier and ``path`` being the path as defined in the asset dictionary.

        Assets of the types ``js``, ``css`` and ``less`` will be automatically bundled by OctoPrint using
        `Flask-Assets <http://flask-assets.readthedocs.org/en/latest/>`_.

        :return dict: a dictionary describing the static assets to publish for the plugin
        """
        return {}


class TemplatePlugin(OctoPrintPlugin, ReloadNeedingPlugin):
    """
    Using the ``TemplatePlugin`` mixin plugins may inject their own components into the OctoPrint web interface.

    Currently OctoPrint supports the following types of injections out of the box:

    Navbar
       The right part of the navigation bar located at the top of the UI can be enriched with additional links. Note that
       with the current implementation, plugins will always be located *to the left* of the existing links.

       The included template must be called ``<plugin identifier>_navbar.jinja2`` (e.g. ``myplugin_navbar.jinja2``) unless
       overridden by the configuration supplied through :func:`get_template_configs`.

       The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
       wrapper structure will have all additional classes and styles applied as specified via the configuration supplied
       through :func:`get_template_configs`.

    Sidebar
       The left side bar containing Connection, State and Files sections can be enriched with additional sections. Note
       that with the current implementations, plugins will always be located *beneath* the existing sections.

       The included template must be called ``<plugin identifier>_sidebar.jinja2`` (e.g. ``myplugin_sidebar.jinja2``) unless
       overridden by the configuration supplied through :func:`get_template_configs`.

       The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
       wrapper divs for both the whole box as well as the content pane will have all additional classes and styles applied
       as specified via the configuration supplied through :func:`get_template_configs`.

    Tabs
       The available tabs of the main part of the interface may be extended with additional tabs originating from within
       plugins. Note that with the current implementation, plugins will always be located *to the right* of the existing
       tabs.

       The included template must be called ``<plugin identifier>_tab.jinja2`` (e.g. ``myplugin_tab.jinja2``) unless
       overridden by the configuration supplied through :func:`get_template_configs`.

       The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
       wrapper div and the link in the navigation will have the additional classes and styles applied as specified via the
       configuration supplied through :func:`get_template_configs`.

    Settings
       Plugins may inject a dialog into the existing settings view. Note that with the current implementation, plugins
       will always be listed beneath the "Plugins" header in the settings link list, ordered alphabetically after
       their displayed name.

       The included template must be called ``<plugin identifier>_settings.jinja2`` (e.g. ``myplugin_settings.jinja2``) unless
       overridden by the configuration supplied through :func:`get_template_configs`.

       The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
       wrapper div and the link in the navigation will have the additional classes and styles applied as defined via the
       configuration through :func:`get_template_configs`.

    Wizards
       Plugins may define wizard dialogs to display to the user if necessary (e.g. in case of missing information that
       needs to be queried from the user to make the plugin work). Note that with the current implementation, all
       wizard dialogs will be will always be sorted by their ``mandatory`` attribute (which defaults to ``False``) and then
       alphabetically by their ``name``. Hence, mandatory wizard steps will come first, sorted alphabetically, then the
       optional steps will follow, also alphabetically. A wizard dialog provided through a plugin will only be displayed
       if the plugin reports the wizard as being required through :meth:`~octoprint.plugin.WizardPlugin.is_wizard_required`.
       Please also refer to the :class:`~octoprint.plugin.WizardPlugin` mixin for further details on this.

       The included template must be called ``<plugin identifier>_wizard.jinja2`` (e.g. ``myplugin_wizard.jinja2``) unless
       overridden by the configuration supplied through :func:`get_template_configs`.

       The template will be already wrapped into the necessary structure, plugins just need to supply the pure content.
       The wrapper div and the link in the wizard navigation will have the additional classes and styles applied as defined
       via the configuration supplied through :func:`get_template_configs`.

       .. note::

          A note about ``mandatory`` wizard steps: In the current implementation, marking a wizard step as
          mandatory will *only* make it styled accordingly. It is the task of the :ref:`view model <sec-plugins-viewmodels>`
          to actually prevent the user from skipping the dialog by implementing the ``onWizardTabChange``
          callback and returning ``false`` there if it is detected that the user hasn't yet filled in the
          wizard step.

       .. versionadded:: 1.3.0

    About
       Plugins may define additional panels into OctoPrint's "About" dialog. Note that with the current implementation
       further about dialog panels will be sorted alphabetically by their name and sorted after the predefined ones.

       The included template must be called ``<plugin identifier>_about.jinja2`` (e.g. ``myplugin_about.jinja2``) unless
       overridden by the configuration supplied through :func:`get_template_configs`.

       The template will be already wrapped into the necessary structure, plugins just need to supply the pure content. The
       wrapped div and the link in the navigation will have the additional classes and styles applied as defined via
       the configuration supplied through :func:`get_template_configs`.

       .. versionadded:: 1.3.0

    Generic
       Plugins may also inject arbitrary templates into the page of the web interface itself, e.g. in order to
       add overlays or dialogs to be called from within the plugin's JavaScript code.

    .. figure:: ../images/template-plugin-types-main.png
       :align: center
       :alt: Template injection types in the main part of the interface

       Template injection types in the main part of the interface

    .. figure:: ../images/template-plugin-types-settings.png
       :align: center
       :alt: Template injection types in the settings

       Template injection types in the settings

    You can find an example for a simple plugin which injects navbar, tab and settings content into the interface in
    the "helloworld" plugin in OctoPrint's :ref:`Plugin Tutorial <sec-plugins-gettingstarted>`.

    Plugins may replace existing components, see the ``replaces`` keyword in the template configurations returned by
    :meth:`.get_template_configs` below. Note that if a plugin replaces a core component, it is the plugin's
    responsibility to ensure that all core functionality is still maintained.

    Plugins can also add additional template types by implementing the :ref:`octoprint.ui.web.templatetypes <sec-plugins-hook-ui-web-templatetypes>` hook.

    ``TemplatePlugin`` is a :class:`~octoprint.plugin.core.ReloadNeedingPlugin`.
    """

    @property
    def template_folder_key(self):
        return f"plugin_{self._identifier}"

    def get_template_configs(self):
        """
        Allows configuration of injected navbar, sidebar, tab and settings templates (and also additional templates of
        types specified by plugins through the :ref:`octoprint.ui.web.templatetypes <sec-plugins-hook-ui-web-templatetypes>` hook).
        Should be a list containing one configuration object per template to inject. Each configuration object is
        represented by a dictionary which may contain the following keys:

        .. list-table::
           :widths: 5 95

           * - type
             - The template type the configuration is targeting. Possible values here are ``navbar``, ``sidebar``,
               ``tab``, ``settings`` and ``generic``. Mandatory.
           * - name
             - The name of the component, if not set the name of the plugin will be used. The name will be visible at
               a location depending on the ``type``:

                 * ``navbar``: unused
                 * ``sidebar``: sidebar heading
                 * ``tab``: tab heading
                 * ``settings``: settings link
                 * ``wizard``: wizard link
                 * ``about``: about link
                 * ``generic``: unused

           * - template
             - Name of the template to inject, default value depends on the ``type``:

                 * ``navbar``: ``<plugin identifier>_navbar.jinja2``
                 * ``sidebar``: ``<plugin identifier>_sidebar.jinja2``
                 * ``tab``: ``<plugin identifier>_tab.jinja2``
                 * ``settings``: ``<plugin identifier>_settings.jinja2``
                 * ``wizard``: ``<plugin identifier>_wizard.jinja2``
                 * ``about``: ``<plugin identifier>_about.jinja2``
                 * ``generic``: ``<plugin identifier>.jinja2``

           * - suffix
             - Suffix to attach to the component identifier and the div identifier of the injected template. Will be
               ``_<index>`` if not provided and not the first template of the type, with ``index`` counting from 1 and
               increasing for each template of the same type.

               Example: If your plugin with identifier ``myplugin`` defines two tab components like this:

               .. code-block:: python

                  return [
                      dict(type="tab", template="myplugin_first_tab.jinja2"),
                      dict(type="tab", template="myplugin_second_tab.jinja2")
                  ]

               then the first tab will have the component identifier ``plugin_myplugin`` and the second one will have
               the component identifier ``plugin_myplugin_2`` (the generated divs will be ``tab_plugin_myplugin`` and
               ``tab_plugin_myplugin_2`` accordingly). Notice that the first tab is *not* called ``plugin_myplugin_1`` --
               as stated above while the ``index`` used as default suffix starts counting at 1, it will not be applied
               for the first component of a given type.

               If on the other hand your plugin's definition looks like this:

               .. code-block:: python

                  return [
                      dict(type="tab", template="myplugin_first_tab_jinja2", suffix="_1st"),
                      dict(type="tab", template="myplugin_second_tab_jinja2", suffix="_2nd")
                  ]

               then the generated component identifier will be ``plugin_myplugin_1st`` and ``plugin_myplugin_2nd``
               (and the divs will be ``tab_plugin_myplugin_1st`` and ``tab_plugin_myplugin_2nd``).

           * - div
             - Id for the div containing the component. If not provided, defaults to ``<type>_plugin_<plugin identifier>`` plus
               the ``suffix`` if provided or required.
           * - replaces
             - Id of the component this one replaces, might be either one of the core components or a component
               provided by another plugin. A list of the core component identifiers can be found
               :ref:`in the configuration documentation <sec-configuration-config_yaml-appearance>`. The identifiers of
               other plugin components always follow the format described above.
           * - custom_bindings
             - A boolean value indicating whether the default view model should be bound to the component (``false``)
               or if a custom binding will be used by the plugin (``true``, default).
           * - data_bind
             - Additional knockout data bindings to apply to the component, can be used to add further behaviour to
               the container based on internal state if necessary.
           * - classes
             - Additional classes to apply to the component, as a list of individual classes
               (e.g. ``classes=["myclass", "myotherclass"]``) which will be joined into the correct format by the template engine.
           * - styles
             - Additional CSS styles to apply to the component, as a list of individual declarations
               (e.g. ``styles=["color: red", "display: block"]``) which will be joined into the correct format by the template
               engine.

        Further keys to be included in the dictionary depend on the type:

        ``sidebar`` type

           .. list-table::
              :widths: 5 95

              * - icon
                - Icon to use for the sidebar header, should be the name of a Font Awesome icon without the leading ``icon-`` part.
              * - template_header
                - Additional template to include in the head section of the sidebar item. For an example of this, see the additional
                  options included in the "Files" section.
              * - classes_wrapper
                - Like ``classes`` but only applied to the whole wrapper around the sidebar box.
              * - classes_content
                - Like ``classes`` but only applied to the content pane itself.
              * - styles_wrapper
                - Like ``styles`` but only applied to the whole wrapper around the sidebar box.
              * - styles_content
                - Like ``styles`` but only applied to the content pane itself

        ``tab`` type and ``settings`` type

           .. list-table::
              :widths: 5 95

              * - classes_link
                - Like ``classes`` but only applied to the link in the navigation.
              * - classes_content
                - Like ``classes`` but only applied to the content pane itself.
              * - styles_link
                - Like ``styles`` but only applied to the link in the navigation.
              * - styles_content
                - Like ``styles`` but only applied to the content pane itself.

        ``wizard`` type

           .. list-table::
              :widths: 5 95

              * - mandatory
                - Whether the wizard step is mandatory (True) or not (False). Optional,
                  defaults to False. If set to True, OctoPrint will sort visually mark
                  the step as mandatory in the UI (bold in the navigation and a little
                  alert) and also sort it into the first half.

        .. note::

           As already outlined above, each template type has a default template name (i.e. the default navbar template
           of a plugin is called ``<plugin identifier>_navbar.jinja2``), which may be overridden using the template configuration.
           If a plugin needs to include more than one template of a given type, it needs to provide an entry for each of
           those, since the implicit default template will only be included automatically if no other templates of that
           type are defined.

           Example: If you have a plugin that injects two tab components, one defined in the template file
           ``myplugin_tab.jinja2`` (the default template) and one in the template ``myplugin_othertab.jinja2``, you
           might be tempted to just return the following configuration since one your templates is named by the default
           template name:

           .. code-block:: python

              return [
                  dict(type="tab", template="myplugin_othertab.jinja2")
              ]

           This will only include the tab defined in ``myplugin_othertab.jinja2`` though, ``myplugin_tab.jinja2`` will
           not be included automatically since the presence of a definition for the ``tab`` type overrides the automatic
           injection of the default template. You'll have to include it explicitly:

           .. code-block:: python

              return [
                  dict(type="tab", template="myplugin_tab.jinja2"),
                  dict(type="tab", template="myplugin_othertab.jinja2")
              ]

        :return list: a list containing the configuration options for the plugin's injected templates
        """
        return []

    def get_template_vars(self):
        """
        Defines additional template variables to include into the template renderer. Variable names will be prefixed
        with ``plugin_<plugin identifier>_``.

        :return dict: a dictionary containing any additional template variables to include in the renderer
        """
        return {}

    def get_template_folder(self):
        """
        Defines the folder where the plugin stores its templates. Override this if your plugin stores its templates at
        some other place than the ``templates`` sub folder in the plugin base directory.

        :return string: the absolute path to the folder where the plugin stores its jinja2 templates
        """
        import os

        return os.path.join(self._basefolder, "templates")


class UiPlugin(OctoPrintPlugin, SortablePlugin):
    """
    The ``UiPlugin`` mixin allows plugins to completely replace the UI served
    by OctoPrint when requesting the main page hosted at `/`.

    OctoPrint will query whether your mixin implementation will handle a
    provided request by calling :meth:`~octoprint.plugin.UiPlugin.will_handle_ui` with the Flask
    `Request <http://flask.pocoo.org/docs/0.10/api/#flask.Request>`_ object as
    parameter. If you plugin returns `True` here, OctoPrint will next call
    :meth:`~octoprint.plugin.UiPlugin.on_ui_render` with a few parameters like
    - again - the Flask Request object and the render keyword arguments as
    used by the default OctoPrint web interface. For more information see below.

    There are two methods used in order to allow for caching of the actual
    response sent to the client. Whatever a plugin implementation returns
    from the call to its :meth:`~octoprint.plugin.UiPlugin.on_ui_render` method
    will be cached server side. The cache will be emptied in case of explicit
    no-cache headers sent by the client, or if the ``_refresh`` query parameter
    on the request exists and is set to ``true``. To prevent caching of the
    response altogether, a plugin may set no-cache headers on the returned
    response as well.

    ``UiPlugin`` is a :class:`~octoprint.plugin.core.SortablePlugin` with a sorting context
    for :meth:`~octoprint.plugin.UiPlugin.will_handle_ui`. The first plugin to return ``True``
    for :meth:`~octoprint.plugin.UiPlugin.will_handle_ui` will be the one whose ui will be used,
    no further calls to :meth:`~octoprint.plugin.UiPlugin.on_ui_render` will be performed.

    If implementations want to serve custom templates in the :meth:`~octoprint.plugin.UiPlugin.on_ui_render`
    method it is recommended to also implement the :class:`~octoprint.plugin.TemplatePlugin`
    mixin.

    **Example**

    What follows is a very simple example that renders a different (non functional and
    only exemplary) UI if the requesting client has a UserAgent string hinting
    at it being a mobile device:

    .. onlineinclude:: https://raw.githubusercontent.com/OctoPrint/Plugin-Examples/master/dummy_mobile_ui/__init__.py
       :linenos:
       :tab-width: 4
       :caption: `dummy_mobile_ui/__init__.py <https://github.com/OctoPrint/Plugin-Examples/blob/master/dummy_mobile_ui/__init__.py>`_

    .. onlineinclude:: https://raw.githubusercontent.com/OctoPrint/Plugin-Examples/master/dummy_mobile_ui/templates/dummy_mobile_ui_index.jinja2
       :linenos:
       :tab-width: 4
       :caption: `dummy_mobile_ui/templates/dummy_mobile_ui_index.jinja2 <https://github.com/OctoPrint/Plugin-Examples/blob/master/dummy_mobile_ui/templates/dummy_mobile_ui_index.jinja2>`_

    Try installing the above plugin ``dummy_mobile_ui`` (also available in the
    `plugin examples repository <https://github.com/OctoPrint/Plugin-Examples/blob/master/dummy_mobile_ui>`_)
    into your OctoPrint instance. If you access it from a regular desktop browser,
    you should still see the default UI. However if you access it from a mobile
    device (make sure to not have that request the desktop version of pages!)
    you should see the very simple dummy page defined above.

    **Preemptive and Runtime Caching**

    OctoPrint will also cache your custom UI for you in its server side UI cache, making sure
    it only gets re-rendered if the request demands that (by having no-cache headers set) or if
    the cache gets invalidated otherwise.

    In order to be able to do that, the ``UiPlugin`` offers overriding some cache specific
    methods used for figuring out the source files whose modification time to use for cache invalidation
    as well as override possibilities for ETag and LastModified calculation. Additionally there are
    methods to allow persisting call parameters to allow for preemptively caching your UI during
    server startup (basically eager caching instead of lazily waiting for the first request).

    See below for details on this.

    .. versionadded:: 1.3.0
    """

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def will_handle_ui(self, request):
        """
        Called by OctoPrint to determine if the mixin implementation will be
        able to handle the ``request`` provided as a parameter.

        Return ``True`` here to signal that your implementation will handle
        the request and that the result of its :meth:`~octoprint.plugin.UiPlugin.on_ui_render` method
        is what should be served to the user.

        The execution order of calls to this method can be influenced via the sorting context
        ``UiPlugin.will_handle_ui``.

        Arguments:
            request (flask.Request): A Flask `Request <http://flask.pocoo.org/docs/0.10/api/#flask.Request>`_
                object.

        Returns:
            bool: ``True`` if the implementation will serve the request,
                ``False`` otherwise.
        """
        return False

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def on_ui_render(self, now, request, render_kwargs):
        """
        Called by OctoPrint to retrieve the response to send to the client
        for the ``request`` to ``/``. Only called if :meth:`~octoprint.plugin.UiPlugin.will_handle_ui`
        returned ``True``.

        ``render_kwargs`` will be a dictionary (whose contents are cached) which
        will contain the following key and value pairs (note that not all
        key value pairs contained in the dictionary are listed here, only
        those you should depend on as a plugin developer at the current time):

        .. list-table::
           :widths: 5 95

           * - debug
             - ``True`` if debug mode is enabled, ``False`` otherwise.
           * - firstRun
             - ``True`` if the server is being run for the first time (not
               configured yet), ``False`` otherwise.
           * - version
             - OctoPrint's version information. This is a ``dict`` with the
               following keys:

               .. list-table::
                  :widths: 5 95

                  * - number
                    - The version number (e.g. ``x.y.z``)
                  * - branch
                    - The GIT branch from which the OctoPrint instance was built
                      (e.g. ``master``)
                  * - display
                    - The full human readable version string, including the
                      branch information (e.g. ``x.y.z (master branch)``

           * - uiApiKey
             - The UI API key to use for unauthorized API requests. This is
               freshly generated on every server restart.
           * - templates
             - Template data to render in the UI. Will be a ``dict`` containing entries
               for all known template types.

               The sub structure for each key will be as follows:

               .. list-table::
                  :widths: 5 95

                  * - order
                    - A list of template names in the order they should appear
                      in the final rendered page
                  * - entries
                    - The template entry definitions to render. Depending on the
                      template type those are either 2-tuples of a name and a ``dict``
                      or directly ``dicts`` with information regarding the
                      template to render.

                      For the possible contents of the data ``dicts`` see the
                      :class:`~octoprint.plugin.TemplatePlugin` mixin.

           * - pluginNames
             - A list of names of :class:`~octoprint.plugin.TemplatePlugin`
               implementation that were enabled when creating the ``templates``
               value.
           * - locales
             - The locales for which there are translations available.
           * - supportedExtensions
             - The file extensions supported for uploads.

        On top of that all additional template variables as provided by :meth:`~octoprint.plugin.TemplatePlugin.get_template_vars`
        will be contained in the dictionary as well.

        Arguments:
            now (datetime.datetime): The datetime instance representing "now"
                for this request, in case your plugin implementation needs this
                information.
            request (flask.Request): A Flask `Request <http://flask.pocoo.org/docs/0.10/api/#flask.Request>`_ object.
            render_kwargs (dict): The (cached) render keyword arguments that
                would usually be provided to the core UI render function.

        Returns:
            flask.Response: Should return a Flask `Response <http://flask.pocoo.org/docs/0.10/api/#flask.Response>`_
                object that can be served to the requesting client directly. May be
                created with ``flask.make_response`` combined with something like
                ``flask.render_template``.
        """

        return None

    # noinspection PyMethodMayBeStatic
    def get_ui_additional_key_data_for_cache(self):
        """
        Allows to return additional data to use in the cache key.

        Returns:
            list, tuple: A list or tuple of strings to use in the cache key. Will be joined by OctoPrint
                using ``:`` as separator and appended to the existing ``ui:<identifier>:<base url>:<locale>``
                cache key. Ignored if ``None`` is returned.

        .. versionadded:: 1.3.0
        """
        return None

    # noinspection PyMethodMayBeStatic
    def get_ui_additional_tracked_files(self):
        """
        Allows to return additional files to track for validating existing caches. By default OctoPrint
        will track all declared templates, assets and translation files in the system. Additional
        files can be added by a plugin through this callback.

        Returns:
            list: A list of paths to additional files whose modification to track for (in)validating
                the cache. Ignored if ``None`` is returned.

        .. versionadded:: 1.3.0
        """
        return None

    # noinspection PyMethodMayBeStatic
    def get_ui_custom_tracked_files(self):
        """
        Allows to define a complete separate set of files to track for (in)validating the cache. If this
        method returns something, the templates, assets and translation files won't be tracked, only the
        files specified in the returned list.

        Returns:
            list: A list of paths representing the only files whose modification to track for (in)validating
                the cache. Ignored if ``None`` is returned.

        .. versionadded:: 1.3.0
        """
        return None

    # noinspection PyMethodMayBeStatic
    def get_ui_custom_etag(self):
        """
        Allows to use a custom way to calculate the ETag, instead of the default method (hashing
        OctoPrint's version, tracked file paths and ``LastModified`` value).

        Returns:
            str: An alternatively calculated ETag value. Ignored if ``None`` is returned (default).

        .. versionadded:: 1.3.0
        """
        return None

    # noinspection PyMethodMayBeStatic
    def get_ui_additional_etag(self, default_additional):
        """
        Allows to provide a list of additional fields to use for ETag generation.

        By default the same list will be returned that is also used in the stock UI (and injected
        via the parameter ``default_additional``).

        Arguments:
            default_additional (list): The list of default fields added to the ETag of the default UI

        Returns:
            (list): A list of additional fields for the ETag generation, or None

        .. versionadded:: 1.3.0
        """
        return default_additional

    # noinspection PyMethodMayBeStatic
    def get_ui_custom_lastmodified(self):
        """
        Allows to calculate the LastModified differently than using the most recent modification
        date of all tracked files.

        Returns:
            int: An alternatively calculated LastModified value. Ignored if ``None`` is returned (default).

        .. versionadded:: 1.3.0
        """
        return None

    # noinspection PyMethodMayBeStatic
    def get_ui_preemptive_caching_enabled(self):
        """
        Allows to control whether the view provided by the plugin should be preemptively
        cached on server startup (default) or not.

        Have this return False if you do not want your plugin's UI to ever be preemptively cached.

        Returns:
            bool: Whether to enable preemptive caching (True, default) or not (False)
        """
        return True

    # noinspection PyMethodMayBeStatic
    def get_ui_data_for_preemptive_caching(self):
        """
        Allows defining additional data to be persisted in the preemptive cache configuration, on
        top of the request path, base URL and used locale.

        Returns:
            dict: Additional data to persist in the preemptive cache configuration.

        .. versionadded:: 1.3.0
        """
        return None

    # noinspection PyMethodMayBeStatic
    def get_ui_additional_request_data_for_preemptive_caching(self):
        """
        Allows defining additional request data to persist in the preemptive cache configuration and
        to use for the fake request used for populating the preemptive cache.

        Keys and values are used as keyword arguments for creating the
        `Werkzeug EnvironBuilder <http://werkzeug.pocoo.org/docs/0.11/test/#werkzeug.test.EnvironBuilder>`_
        used for creating the fake request.

        Returns:
            dict: Additional request data to persist in the preemptive cache configuration and to
                use for request environment construction.

        .. versionadded:: 1.3.0
        """
        return None

    # noinspection PyMethodMayBeStatic
    def get_ui_preemptive_caching_additional_unless(self):
        """
        Allows defining additional reasons for temporarily not adding a preemptive cache record for
        your plugin's UI.

        OctoPrint will call this method when processing a UI request, to determine whether to record the
        access or not. If you return ``True`` here, no record will be created.

        Returns:
            bool: Whether to suppress a record (True) or not (False, default)

        .. versionadded:: 1.3.0
        """
        return False

    # noinspection PyMethodMayBeStatic
    def get_ui_custom_template_filter(self, default_template_filter):
        """
        Allows to specify a custom template filter to use for filtering the template contained in the
        ``render_kwargs`` provided to the templating sub system.

        Only relevant for UiPlugins that actually utilize the stock templates of OctoPrint.

        By default simply returns the provided ``default_template_filter``.

        Arguments:
            default_template_filter (callable): The default template filter used by the default UI

        Returns:
            (callable) A filter function accepting the ``template_type`` and ``template_key`` of a template
            and returning ``True`` to keep it and ``False`` to filter it out. If ``None`` is returned, no
            filtering will take place.

        .. versionadded:: 1.3.0
        """
        return default_template_filter

    # noinspection PyMethodMayBeStatic
    def get_ui_permissions(self):
        """
        Determines a list of permissions that need to be on the current user session. If
        these requirements are not met, OctoPrint will instead redirect to a login
        screen.

        Plugins may override this with their own set of permissions. Returning an empty
        list will instruct OctoPrint to never show a login dialog when this UiPlugin's
        view renders, in which case it will fall to your plugin to implement its own
        login logic.

        Returns:
            (list) A list of permissions which to check the current user session against.
            May be empty to indicate that no permission checks should be made by OctoPrint.

        .. versionadded: 1.5.0
        """
        from octoprint.access.permissions import Permissions

        return [Permissions.STATUS, Permissions.SETTINGS_READ]


class WizardPlugin(OctoPrintPlugin, ReloadNeedingPlugin):
    """
    The ``WizardPlugin`` mixin allows plugins to report to OctoPrint whether
    the ``wizard`` templates they define via the :class:`~octoprint.plugin.TemplatePlugin`
    should be displayed to the user, what details to provide to their respective
    wizard frontend components and what to do when the wizard is finished
    by the user.

    OctoPrint will only display such wizard dialogs to the user which belong
    to plugins that

      * report ``True`` in their :func:`is_wizard_required` method and
      * have not yet been shown to the user in the version currently being reported
        by the :meth:`~octoprint.plugin.WizardPlugin.get_wizard_version` method

    Example: If a plugin with the identifier ``myplugin`` has a specific
    setting ``some_key`` it needs to have filled by the user in order to be
    able to work at all, it would probably test for that setting's value in
    the :meth:`~octoprint.plugin.WizardPlugin.is_wizard_required` method and
    return ``True`` if the value is unset:

    .. code-block:: python

       class MyPlugin(octoprint.plugin.SettingsPlugin,
                      octoprint.plugin.TemplatePlugin,
                      octoprint.plugin.WizardPlugin):

           def get_default_settings(self):
               return dict(some_key=None)

           def is_wizard_required(self):
               return self._settings.get(["some_key"]) is None

    OctoPrint will then display the wizard dialog provided by the plugin through
    the :class:`TemplatePlugin` mixin. Once the user finishes the wizard on the
    frontend, OctoPrint will store that it already showed the wizard of ``myplugin``
    in the version reported by :meth:`~octoprint.plugin.WizardPlugin.get_wizard_version`
    - here ``None`` since that is the default value returned by that function
    and the plugin did not override it.

    If the plugin in a later version needs another setting from the user in order
    to function, it will also need to change the reported version in order to
    have OctoPrint reshow the dialog. E.g.

    .. code-block:: python

       class MyPlugin(octoprint.plugin.SettingsPlugin,
                      octoprint.plugin.TemplatePlugin,
                      octoprint.plugin.WizardPlugin):

           def get_default_settings(self):
               return dict(some_key=None, some_other_key=None)

           def is_wizard_required(self):
               some_key_unset = self._settings.get(["some_key"]) is None
               some_other_key_unset = self._settings.get(["some_other_key"]) is None

               return some_key_unset or some_other_key_unset

           def get_wizard_version(self):
               return 1

    ``WizardPlugin`` is a :class:`~octoprint.plugin.core.ReloadNeedingPlugin`.
    """

    # noinspection PyMethodMayBeStatic
    def is_wizard_required(self):
        """
        Allows the plugin to report whether it needs to display a wizard to the
        user or not.

        Defaults to ``False``.

        OctoPrint will only include those wizards from plugins which are reporting
        their wizards as being required through this method returning ``True``.
        Still, if OctoPrint already displayed that wizard in the same version
        to the user once it won't be displayed again regardless whether this
        method returns ``True`` or not.
        """
        return False

    # noinspection PyMethodMayBeStatic
    def get_wizard_version(self):
        """
        The version of this plugin's wizard. OctoPrint will only display a wizard
        of the same plugin and wizard version once to the user. After they
        finish the wizard, OctoPrint will remember that it already showed this
        wizard in this particular version and not reshow it.

        If a plugin needs to show its wizard to the user again (e.g. because
        of changes in the required settings), increasing this value is the
        way to notify OctoPrint of these changes.

        Returns:
            int or None: an int signifying the current wizard version, should be incremented by plugins whenever there
                         are changes to the plugin that might necessitate reshowing the wizard if it is required. ``None``
                         will also be accepted and lead to the wizard always be ignored unless it has never been finished
                         so far
        """
        return None

    # noinspection PyMethodMayBeStatic
    def get_wizard_details(self):
        """
        Called by OctoPrint when the wizard wrapper dialog is shown. Allows the plugin to return data
        that will then be made available to the view models via the view model callback ``onWizardDetails``.

        Use this if your plugin's view model that handles your wizard dialog needs additional
        data to perform its task.

        Returns:
            dict: a dictionary containing additional data to provide to the frontend. Whatever the plugin
                  returns here will be made available on the wizard API under the plugin's identifier
        """
        return {}

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def on_wizard_finish(self, handled):
        """
        Called by OctoPrint whenever the user finishes a wizard session.

        The ``handled`` parameter will indicate whether that plugin's wizard was
        included in the wizard dialog presented to the user (so the plugin providing
        it was reporting that the wizard was required and the wizard plus version was not
        ignored/had already been seen).

        Use this to do any clean up tasks necessary after wizard completion.

        Arguments:
            handled (bool): True if the plugin's wizard was previously reported as
                            required, not ignored and thus presented to the user,
                            False otherwise
        """
        pass

    # noinspection PyProtectedMember
    @classmethod
    def is_wizard_ignored(cls, seen_wizards, implementation):
        """
        Determines whether the provided implementation is ignored based on the
        provided information about already seen wizards and their versions or not.

        A wizard is ignored if

          * the current and seen versions are identical
          * the current version is None and the seen version is not
          * the seen version is not None and the current version is less or equal than the seen one

        .. code-block:: none

               |  current  |
               | N | 1 | 2 |   N = None
           ----+---+---+---+   X = ignored
           s N | X |   |   |
           e --+---+---+---+
           e 1 | X | X |   |
           n --+---+---+---+
             2 | X | X | X |
           ----+---+---+---+

        Arguments:
            seen_wizards (dict): A dictionary with information about already seen
                wizards and their versions. Mappings from the identifiers of
                the plugin providing the wizard to the reported wizard
                version (int or None) that was already seen by the user.
            implementation (object): The plugin implementation to check.

        Returns:
            bool: False if the provided ``implementation`` is either not a :class:`WizardPlugin`
                  or has not yet been seen (in this version), True otherwise
        """

        if not isinstance(implementation, cls):
            return False

        name = implementation._identifier
        if name not in seen_wizards:
            return False

        seen = seen_wizards[name]
        wizard_version = implementation.get_wizard_version()

        current = None
        if wizard_version is not None:
            try:
                current = int(wizard_version)
            except ValueError as e:
                import logging

                logging.getLogger(__name__).log(
                    "WizardPlugin {} returned invalid value {} for wizard version: {}".format(
                        name, wizard_version, str(e)
                    )
                )

        return (
            (current == seen)
            or (current is None and seen is not None)
            or (seen is not None and current <= seen)
        )


class SimpleApiPlugin(OctoPrintPlugin):
    """
    Utilizing the ``SimpleApiPlugin`` mixin plugins may implement a simple API based around one GET resource and one
    resource accepting JSON commands POSTed to it. This is the easy alternative for plugin's which don't need the
    full power of a `Flask Blueprint <http://flask.pocoo.org/docs/0.10/blueprints/>`_ that the :class:`BlueprintPlugin`
    mixin offers.

    Use this mixin if all you need to do is return some kind of dynamic data to your plugin from the backend
    and/or want to react to simple commands which boil down to a type of command and a few flat parameters
    supplied with it.

    The simple API constructed by OctoPrint for you will be made available under ``/api/plugin/<plugin identifier>/``.
    OctoPrint will do some preliminary request validation for your defined commands, making sure the request body is in
    the correct format (content type must be JSON) and contains all obligatory parameters for your command.

    Let's take a look at a small example for such a simple API and how you would go about calling it.

    Take this example of a plugin registered under plugin identifier ``mysimpleapiplugin``:

    .. code-block:: python
       :linenos:

       import octoprint.plugin

       import flask

       class MySimpleApiPlugin(octoprint.plugin.SimpleApiPlugin):
           def get_api_commands(self):
               return dict(
                   command1=[],
                   command2=["some_parameter"]
               )

           def on_api_command(self, command, data):
               import flask
               if command == "command1":
                   parameter = "unset"
                   if "parameter" in data:
                       parameter = "set"
                   self._logger.info("command1 called, parameter is {parameter}".format(**locals()))
               elif command == "command2":
                   self._logger.info("command2 called, some_parameter is {some_parameter}".format(**data))

           def on_api_get(self, request):
               return flask.jsonify(foo="bar")

       __plugin_implementation__ = MySimpleApiPlugin()

    Our plugin defines two commands, ``command1`` with no mandatory parameters and ``command2`` with one
    mandatory parameter ``some_parameter``.

    ``command1`` can also accept an optional parameter ``parameter``, and will log whether
    that parameter was set or unset. ``command2`` will log the content of the mandatory ``some_parameter`` parameter.

    A valid POST request for ``command2`` sent to ``/api/plugin/mysimpleapiplugin`` would look like this:

    .. sourcecode:: http

       POST /api/plugin/mysimpleapiplugin HTTP/1.1
       Host: example.com
       Content-Type: application/json
       X-Api-Key: abcdef...

       {
         "command": "command2",
         "some_parameter": "some_value",
         "some_optional_parameter": 2342
       }

    which would produce a response like this:

    .. sourcecode:: http

       HTTP/1.1 204 No Content

    and print something like this line to ``octoprint.log``::

       2015-02-12 17:40:21,140 - octoprint.plugins.mysimpleapiplugin - INFO - command2 called, some_parameter is some_value

    A GET request on our plugin's simple API resource will only return a JSON document like this:

    .. sourcecode:: http

       HTTP/1.1 200 Ok
       Content-Type: application/json

       {
         "foo": "bar"
       }
    """

    # noinspection PyMethodMayBeStatic
    def get_api_commands(self):
        """
        Return a dictionary here with the keys representing the accepted commands and the values being lists of
        mandatory parameter names.
        """
        return None

    # noinspection PyMethodMayBeStatic
    def is_api_adminonly(self):
        """
        Return True if the API is only available to users having the admin role.
        """
        return False

    # noinspection PyMethodMayBeStatic
    def on_api_command(self, command, data):
        """
        Called by OctoPrint upon a POST request to ``/api/plugin/<plugin identifier>``. ``command`` will contain one of
        the commands as specified via :func:`get_api_commands`, ``data`` will contain the full request body parsed
        from JSON into a Python dictionary. Note that this will also contain the ``command`` attribute itself. For the
        example given above, for the ``command2`` request the ``data`` received by the plugin would be equal to
        ``dict(command="command2", some_parameter="some_value")``.

        If your plugin returns nothing here, OctoPrint will return an empty response with return code ``204 No content``
        for you. You may also return regular responses as you would return from any Flask view here though, e.g.
        ``return flask.jsonify(result="some json result")`` or ``flask.abort(404)``.

        :param string command: the command with which the resource was called
        :param dict data:      the full request body of the POST request parsed from JSON into a Python dictionary
        :return: ``None`` in which case OctoPrint will generate a ``204 No content`` response with empty body, or optionally
                 a proper Flask response.
        """
        return None

    # noinspection PyMethodMayBeStatic
    def on_api_get(self, request):
        """
        Called by OctoPrint upon a GET request to ``/api/plugin/<plugin identifier>``. ``request`` will contain the
        received `Flask request object <http://flask.pocoo.org/docs/0.9/api/#flask.Request>`_ which you may evaluate
        for additional arguments supplied with the request.

        If your plugin returns nothing here, OctoPrint will return an empty response with return code ``204 No content``
        for you. You may also return regular responses as you would return from any Flask view here though, e.g.
        ``return flask.jsonify(result="some json result")`` or ``flask.abort(404)``.

        :param request: the Flask request object
        :return: ``None`` in which case OctoPrint will generate a ``204 No content`` response with empty body, or optionally
                 a proper Flask response.
        """
        return None


class BlueprintPlugin(OctoPrintPlugin, RestartNeedingPlugin):
    """
    The ``BlueprintPlugin`` mixin allows plugins to define their own full fledged endpoints for whatever purpose,
    be it a more sophisticated API than what is possible via the :class:`SimpleApiPlugin` or a custom web frontend.

    The mechanism at work here is `Flask's <http://flask.pocoo.org/>`_ own `Blueprint mechanism <http://flask.pocoo.org/docs/0.10/blueprints/>`_.

    The mixin automatically creates a blueprint for you that will be registered under ``/plugin/<plugin identifier>/``.
    All you need to do is decorate all of your view functions with the :func:`route` decorator,
    which behaves exactly the same like Flask's regular ``route`` decorators. Example:

    .. code-block:: python
       :linenos:

       import octoprint.plugin
       import flask

       class MyBlueprintPlugin(octoprint.plugin.BlueprintPlugin):
           @octoprint.plugin.BlueprintPlugin.route("/echo", methods=["GET"])
           def myEcho(self):
               if not "text" in flask.request.values:
                   abort(400, description="Expected a text to echo back.")
               return flask.request.values["text"]

       __plugin_implementation__ = MyBlueprintPlugin()

    Your blueprint will be published by OctoPrint under the base URL ``/plugin/<plugin identifier>/``, so the above
    example of a plugin with the identifier "myblueprintplugin" would be reachable under
    ``/plugin/myblueprintplugin/echo``.

    Just like with regular blueprints you'll be able to create URLs via ``url_for``, just use the prefix
    ``plugin.<plugin identifier>.<method_name>``, e.g.:

    .. code-block:: python

       flask.url_for("plugin.myblueprintplugin.myEcho") # will return "/plugin/myblueprintplugin/echo"

    .. warning::

       As of OctoPrint 1.8.3, endpoints provided through a ``BlueprintPlugin`` do **not** automatically fall under
       OctoPrint's :ref:`CSRF protection <sec-api-general-csrf>`, for reasons of backwards compatibility. There will be a short grace period before this changes. You
       can and should however already opt into CSRF protection for your endpoints by implementing ``is_blueprint_csrf_protected``
       and returning ``True`` from it. You can exempt certain endpoints from CSRF protection by decorating them with
       ``@octoprint.plugin.BlueprintPlugin.csrf_exempt``.

       .. code-block:: python

          class MyPlugin(octoprint.plugin.BlueprintPlugin):
              @octoprint.plugin.BlueprintPlugin.route("/hello_world", methods=["GET"])
              def hello_world(self):
                  # This is a GET request and thus not subject to CSRF protection
                  return "Hello world!"

              @octoprint.plugin.BlueprintPlugin.route("/hello_you", methods=["POST"])
              def hello_you(self):
                  # This is a POST request and thus subject to CSRF protection. It is not exempt.
                  return "Hello you!"

              @octoprint.plugin.BlueprintPlugin.route("/hello_me", methods=["POST"])
              @octoprint.plugin.BlueprintPlugin.csrf_exempt()
              def hello_me(self):
                  # This is a POST request and thus subject to CSRF protection, but this one is exempt.
                  return "Hello me!"

              def is_blueprint_csrf_protected(self):
                  return True

    ``BlueprintPlugin`` implements :class:`~octoprint.plugins.core.RestartNeedingPlugin`.

    .. versionchanged:: 1.8.3
    """

    @staticmethod
    def route(rule, **options):
        """
        A decorator to mark view methods in your BlueprintPlugin subclass. Works just the same as Flask's
        own ``route`` decorator available on blueprints.

        See `the documentation for flask.Blueprint.route <http://flask.pocoo.org/docs/0.10/api/#flask.Blueprint.route>`_
        and `the documentation for flask.Flask.route <http://flask.pocoo.org/docs/0.10/api/#flask.Flask.route>`_ for more
        information.
        """

        from collections import defaultdict

        def decorator(f):
            # We attach the decorator parameters directly to the function object, because that's the only place
            # we can access right now.
            # This neat little trick was adapted from the Flask-Classy project: https://pythonhosted.org/Flask-Classy/
            if not hasattr(f, "_blueprint_rules") or f._blueprint_rules is None:
                f._blueprint_rules = defaultdict(list)
            f._blueprint_rules[f.__name__].append((rule, options))
            return f

        return decorator

    @staticmethod
    def errorhandler(code_or_exception):
        """
        A decorator to mark errorhandlings methods in your BlueprintPlugin subclass. Works just the same as Flask's
        own ``errorhandler`` decorator available on blueprints.

        See `the documentation for flask.Blueprint.errorhandler <http://flask.pocoo.org/docs/0.10/api/#flask.Blueprint.errorhandler>`_
        and `the documentation for flask.Flask.errorhandler <http://flask.pocoo.org/docs/0.10/api/#flask.Flask.errorhandler>`_ for more
        information.

        .. versionadded:: 1.3.0
        """
        from collections import defaultdict

        def decorator(f):
            if (
                not hasattr(f, "_blueprint_error_handler")
                or f._blueprint_error_handler is None
            ):
                f._blueprint_error_handler = defaultdict(list)
            f._blueprint_error_handler[f.__name__].append(code_or_exception)
            return f

        return decorator

    @staticmethod
    def csrf_exempt():
        """
        A decorator to mark a view method in your BlueprintPlugin as exempt from :ref:`CSRF protection <sec-api-general-csrf>`. This makes sense
        if you offer an authenticated API for a certain workflow (see e.g. the bundled appkeys plugin) but in most
        cases should not be needed.

        .. versionadded:: 1.8.3
        """

        def decorator(f):
            if (
                not hasattr(f, "_blueprint_csrf_exempt")
                or f._blueprint_csrf_exempt is None
            ):
                f._blueprint_csrf_exempt = set()
            f._blueprint_csrf_exempt.add(f.__name__)
            return f

        return decorator

    # noinspection PyProtectedMember
    def get_blueprint(self):
        """
        Creates and returns the blueprint for your plugin. Override this if you want to define and handle your blueprint yourself.

        This method will only be called once during server initialization.

        :return: the blueprint ready to be registered with Flask
        """

        if hasattr(self, "_blueprint"):
            # if we already constructed the blueprint and hence have it cached,
            # return that instance - we don't want to instance it multiple times
            return self._blueprint

        import flask

        from octoprint.server.util.csrf import add_exempt_view

        kwargs = self.get_blueprint_kwargs()
        blueprint = flask.Blueprint(self._identifier, self._identifier, **kwargs)

        # we now iterate over all members of ourselves and look if we find an attribute
        # that has data originating from one of our decorators - we ignore anything
        # starting with a _ to only handle public stuff
        for member in [member for member in dir(self) if not member.startswith("_")]:
            f = getattr(self, member)

            if hasattr(f, "_blueprint_rules") and member in f._blueprint_rules:
                # this attribute was annotated with our @route decorator
                for blueprint_rule in f._blueprint_rules[member]:
                    rule, options = blueprint_rule
                    endpoint = options.pop("endpoint", f.__name__)
                    blueprint.add_url_rule(rule, endpoint, view_func=f, **options)

                    if (
                        hasattr(f, "_blueprint_csrf_exempt")
                        and member in f._blueprint_csrf_exempt
                    ):
                        add_exempt_view(f"plugin.{self._identifier}.{endpoint}")

            if (
                hasattr(f, "_blueprint_error_handler")
                and member in f._blueprint_error_handler
            ):
                # this attribute was annotated with our @error_handler decorator
                for code_or_exception in f._blueprint_error_handler[member]:
                    blueprint.errorhandler(code_or_exception)(f)

        # cache and return the blueprint object
        self._blueprint = blueprint
        return blueprint

    def get_blueprint_kwargs(self):
        """
        Override this if you want your blueprint constructed with additional options such as ``static_folder``,
        ``template_folder``, etc.

        Defaults to the blueprint's ``static_folder`` and ``template_folder`` to be set to the plugin's basefolder
        plus ``/static`` or respectively ``/templates``, or -- if the plugin also implements :class:`AssetPlugin` and/or
        :class:`TemplatePlugin` -- the paths provided by ``get_asset_folder`` and ``get_template_folder`` respectively.
        """
        import os

        if isinstance(self, AssetPlugin):
            static_folder = self.get_asset_folder()
        else:
            static_folder = os.path.join(self._basefolder, "static")

        if isinstance(self, TemplatePlugin):
            template_folder = self.get_template_folder()
        else:
            template_folder = os.path.join(self._basefolder, "templates")

        return {"static_folder": static_folder, "template_folder": template_folder}

    # noinspection PyMethodMayBeStatic
    def is_blueprint_protected(self):
        """
        Whether a login session by a registered user is needed to access the blueprint's endpoints. Requiring
        a session is the default. Note that this only restricts access to the blueprint's dynamic methods, static files
        are always accessible.

        If you want your blueprint's endpoints to have specific permissions, return ``False`` for this and do your
        permissions checks explicitly.
        """
        return True

    # noinspection PyMethodMayBeStatic
    def is_blueprint_csrf_protected(self):
        """
        Whether a blueprint's endpoints are :ref:`CSRF protected <sec-api-general-csrf>`. For now, this defaults to ``False`` to leave it up to
        plugins to decide which endpoints *should* be protected. Long term, this will default to ``True`` and hence
        enforce protection unless a plugin opts out by returning False here.

        If you do not override this method in your mixin implementation, a warning will be logged to the console
        to alert you of the requirement to make a decision here and to not rely on the default implementation, due to the
        forthcoming change in implemented default behaviour.

        .. versionadded:: 1.8.3
        """
        self._logger.warning(
            "The Blueprint of this plugin is relying on the default implementation of "
            "is_blueprint_csrf_protected (newly added in OctoPrint 1.8.3), which in a future version will "
            "be switched from False to True for security reasons. Plugin authors should ensure they explicitly "
            "declare the CSRF protection status in their BlueprintPlugin mixin implementation. "
            "Recommendation is to enable CSRF protection and exempt views that must not use it with the "
            "octoprint.plugin.BlueprintPlugin.csrf_exempt decorator."
        )
        return False

    # noinspection PyMethodMayBeStatic
    def get_blueprint_api_prefixes(self):
        """
        Return all prefixes of your endpoint that are an API that should be containing JSON only.

        Anything that matches this will generate JSON error messages in case of flask.abort
        calls, instead of the default HTML ones.

        Defaults to all endpoints under the blueprint. Limit this further as needed. E.g.,
        if you only want your endpoints /foo, /foo/1 and /bar to be declared as API,
        return ``["/foo", "/bar"]``. A match will be determined via startswith.
        """
        return [""]


class SettingsPlugin(OctoPrintPlugin):
    """
    Including the ``SettingsPlugin`` mixin allows plugins to store and retrieve their own settings within OctoPrint's
    configuration.

    Plugins including the mixing will get injected an additional property ``self._settings`` which is an instance of
    :class:`PluginSettingsManager` already properly initialized for use by the plugin. In order for the manager to
    know about the available settings structure and default values upon initialization, implementing plugins will need
    to provide a dictionary with the plugin's default settings through overriding the method :func:`get_settings_defaults`.
    The defined structure will then be available to access through the settings manager available as ``self._settings``.

    If your plugin needs to react to the change of specific configuration values on the fly, e.g. to adjust the log level
    of a logger when the user changes a corresponding flag via the settings dialog, you can override the
    :func:`on_settings_save` method and wrap the call to the implementation from the parent class with retrieval of the
    old and the new value and react accordingly.

    Example:

    .. code-block:: python

       import octoprint.plugin

       class MySettingsPlugin(octoprint.plugin.SettingsPlugin, octoprint.plugin.StartupPlugin):
           def get_settings_defaults(self):
               return dict(
                   some_setting="foo",
                   some_value=23,
                   sub=dict(
                       some_flag=True
                   )
               )

           def on_settings_save(self, data):
               old_flag = self._settings.get_boolean(["sub", "some_flag"])

               octoprint.plugin.SettingsPlugin.on_settings_save(self, data)

               new_flag = self._settings.get_boolean(["sub", "some_flag"])
               if old_flag != new_flag:
                   self._logger.info("sub.some_flag changed from {old_flag} to {new_flag}".format(**locals()))

           def on_after_startup(self):
               some_setting = self._settings.get(["some_setting"])
               some_value = self._settings.get_int(["some_value"])
               some_flag = self._settings.get_boolean(["sub", "some_flag"])
               self._logger.info("some_setting = {some_setting}, some_value = {some_value}, sub.some_flag = {some_flag}".format(**locals())

       __plugin_implementation__ = MySettingsPlugin()

    Of course, you are always free to completely override both :func:`on_settings_load` and :func:`on_settings_save` if the
    default implementations do not fit your requirements.


    .. warning::

       Make sure to protect sensitive information stored by your plugin that only logged in administrators (or users)
       should have access to via :meth:`~octoprint.plugin.SettingsPlugin.get_settings_restricted_paths`. OctoPrint will
       return its settings on the REST API even to anonymous clients, but will filter out fields it knows are restricted,
       therefore you **must** make sure that you specify sensitive information accordingly to limit access as required!
    """

    config_version_key = "_config_version"
    """Key of the field in the settings that holds the configuration format version."""

    # noinspection PyMissingConstructor
    def __init__(self):
        self._settings = None
        """
        The :class:`~octoprint.plugin.PluginSettings` instance to use for accessing the plugin's settings. Injected by
        the plugin core system upon initialization of the implementation.
        """

    def on_settings_load(self):
        """
        Loads the settings for the plugin, called by the Settings API view in order to retrieve all settings from
        all plugins. Override this if you want to inject additional settings properties that are not stored within
        OctoPrint's configuration.

        .. note::

           The default implementation will return your plugin's settings as is, so just in the structure and in the types
           that are currently stored in OctoPrint's configuration.

           If you need more granular control here, e.g. over the used data types, you'll need to override this method
           and iterate yourself over all your settings, using the proper retriever methods on the settings manager
           to retrieve the data in the correct format.

           The default implementation will also replace any paths that have been restricted by your plugin through
           :func:`~octoprint.plugin.SettingsPlugin.get_settings_restricted_paths` with either the provided
           default value (if one was provided), an empty dictionary (as fallback for restricted dictionaries), an
           empty list (as fallback for restricted lists) or ``None`` values where necessary.
           Make sure to do your own restriction if you decide to fully overload this method.

        :return: the current settings of the plugin, as a dictionary
        """
        import copy

        from flask_login import current_user

        from octoprint.access.permissions import OctoPrintPermission, Permissions

        data = copy.deepcopy(self._settings.get_all_data(merged=True))
        if self.config_version_key in data:
            del data[self.config_version_key]

        restricted_paths = self.get_settings_restricted_paths()

        # noinspection PyShadowingNames
        def restrict_path_unless(data, path, condition):
            if not path:
                return

            if condition():
                return

            node = data

            if len(path) > 1:
                for entry in path[:-1]:
                    if entry not in node:
                        return
                    node = node[entry]

            key = path[-1]
            default_value_available = False
            default_value = None
            if isinstance(key, (list, tuple)):
                # key, default_value tuple
                key, default_value = key
                default_value_available = True

            if key in node:
                if default_value_available:
                    if callable(default_value):
                        default_value = default_value()
                    node[key] = default_value
                else:
                    if isinstance(node[key], dict):
                        node[key] = {}
                    elif isinstance(node[key], (list, tuple)):
                        node[key] = []
                    else:
                        node[key] = None

        conditions = {
            "user": lambda: current_user is not None and not current_user.is_anonymous,
            "admin": lambda: current_user is not None
            and current_user.has_permission(Permissions.SETTINGS),
            "never": lambda: False,
        }

        for level, paths in restricted_paths.items():
            if isinstance(level, OctoPrintPermission):
                condition = lambda: (
                    current_user is not None and current_user.has_permission(level)
                )
            else:
                condition = conditions.get(level, lambda: False)

            for path in paths:
                restrict_path_unless(data, path, condition)

        return data

    def on_settings_save(self, data):
        """
        Saves the settings for the plugin, called by the Settings API view in order to persist all settings
        from all plugins. Override this if you need to directly react to settings changes or want to extract
        additional settings properties that are not stored within OctoPrint's configuration.

        .. note::

           The default implementation will persist your plugin's settings as is, so just in the structure and in the
           types that were received by the Settings API view. Values identical to the default settings values
           will *not* be persisted.

           If you need more granular control here, e.g. over the used data types, you'll need to override this method
           and iterate yourself over all your settings, retrieving them (if set) from the supplied received ``data``
           and using the proper setter methods on the settings manager to persist the data in the correct format.

        Arguments:
            data (dict): The settings dictionary to be saved for the plugin

        Returns:
            dict: The settings that differed from the defaults and were actually saved.
        """
        import octoprint.util

        # get the current data
        current = self._settings.get_all_data()
        if current is None:
            current = {}

        # merge our new data on top of it
        new_current = octoprint.util.dict_merge(current, data)
        if self.config_version_key in new_current:
            del new_current[self.config_version_key]

        # determine diff dict that contains minimal set of changes against the
        # default settings - we only want to persist that, not everything
        diff = octoprint.util.dict_minimal_mergediff(
            self.get_settings_defaults(), new_current
        )

        version = self.get_settings_version()

        to_persist = dict(diff)
        if version:
            to_persist[self.config_version_key] = version

        if to_persist:
            self._settings.set([], to_persist)
        else:
            self._settings.clean_all_data()

        return diff

    # noinspection PyMethodMayBeStatic
    def get_settings_defaults(self):
        """
        Retrieves the plugin's default settings with which the plugin's settings manager will be initialized.

        Override this in your plugin's implementation and return a dictionary defining your settings data structure
        with included default values.
        """
        return {}

    # noinspection PyMethodMayBeStatic
    def get_settings_restricted_paths(self):
        """
        Retrieves the list of paths in the plugin's settings which be restricted on the REST API.

        Override this in your plugin's implementation to restrict whether a path should only be returned to users with
        certain permissions, or never on the REST API.

        Return a ``dict`` with one of the following keys, mapping to a list of paths (as tuples or lists of
        the path elements) for which to restrict access via the REST API accordingly.

           * An :py:class:`~octoprint.access.permissions.OctoPrintPermission` instance: Paths will only be available on the REST API for users with the permission
           * ``admin``: Paths will only be available on the REST API for users with admin rights (any user with the SETTINGS permission)
           * ``user``: Paths will only be available on the REST API when accessed as a logged in user
           * ``never``: Paths will never be returned on the API

        Example:

        .. code-block:: python

           def get_settings_defaults(self):
               return dict(some=dict(admin_only=dict(path="path", foo="foo"),
                                     user_only=dict(path="path", bar="bar")),
                           another=dict(admin_only=dict(path="path"),
                                        field="field"),
                           path=dict(to=dict(never=dict(return="return"))),
                           the=dict(webcam=dict(data="webcam")))

           def get_settings_restricted_paths(self):
               from octoprint.access.permissions import Permissions
               return {'admin':[["some", "admin_only", "path"], ["another", "admin_only", "path"],],
                       'user':[["some", "user_only", "path"],],
                       'never':[["path", "to", "never", "return"],],
                       Permissions.WEBCAM:[["the", "webcam", "data"],]}

           # this will make the plugin return settings on the REST API like this for an anonymous user
           #
           #     dict(some=dict(admin_only=dict(path=None, foo="foo"),
           #                    user_only=dict(path=None, bar="bar")),
           #          another=dict(admin_only=dict(path=None),
           #                       field="field"),
           #          path=dict(to=dict(never=dict(return=None))),
           #          the=dict(webcam=dict(data=None)))
           #
           # like this for a logged in user without the webcam permission
           #
           #     dict(some=dict(admin_only=dict(path=None, foo="foo"),
           #                    user_only=dict(path="path", bar="bar")),
           #          another=dict(admin_only=dict(path=None),
           #                       field="field"),
           #          path=dict(to=dict(never=dict(return=None))),
           #          the=dict(webcam=dict(data=None)))
           #
           # like this for a logged in user with the webcam permission
           #
           #     dict(some=dict(admin_only=dict(path=None, foo="foo"),
           #                    user_only=dict(path="path", bar="bar")),
           #          another=dict(admin_only=dict(path=None),
           #                       field="field"),
           #          path=dict(to=dict(never=dict(return=None))),
           #          the=dict(webcam=dict(data="webcam")))
           #
           # and like this for an admin user
           #
           #     dict(some=dict(admin_only=dict(path="path", foo="foo"),
           #                    user_only=dict(path="path", bar="bar")),
           #          another=dict(admin_only=dict(path="path"),
           #                       field="field"),
           #          path=dict(to=dict(never=dict(return=None))),
           #          the=dict(webcam=dict(data="webcam")))

        .. versionadded:: 1.2.17
        """
        return {}

    # noinspection PyMethodMayBeStatic
    def get_settings_preprocessors(self):
        """
        Retrieves the plugin's preprocessors to use for preprocessing returned or set values prior to returning/setting
        them.

        The preprocessors should be provided as a dictionary mapping the path of the values to preprocess
        (hierarchically) to a transform function which will get the value to transform as only input and should return
        the transformed value.

        Example:

        .. code-block:: python

           def get_settings_defaults(self):
               return dict(some_key="Some_Value", some_other_key="Some_Value")

           def get_settings_preprocessors(self):
               return dict(some_key=lambda x: x.upper()),        # getter preprocessors
                      dict(some_other_key=lambda x: x.lower())   # setter preprocessors

           def some_method(self):
               # getting the value for "some_key" should turn it to upper case
               assert self._settings.get(["some_key"]) == "SOME_VALUE"

               # the value for "some_other_key" should be left alone
               assert self._settings.get(["some_other_key"] = "Some_Value"

               # setting a value for "some_other_key" should cause the value to first be turned to lower case
               self._settings.set(["some_other_key"], "SOME_OTHER_VALUE")
               assert self._settings.get(["some_other_key"]) == "some_other_value"

        Returns:
            (dict, dict): A tuple consisting of two dictionaries, the first being the plugin's preprocessors for
            getters, the second the preprocessors for setters
        """
        return {}, {}

    # noinspection PyMethodMayBeStatic
    def get_settings_version(self):
        """
        Retrieves the settings format version of the plugin.

        Use this to have OctoPrint trigger your migration function if it detects an outdated settings version in
        config.yaml.

        Returns:
            int or None: an int signifying the current settings format, should be incremented by plugins whenever there
            are backwards incompatible changes. Returning None here disables the version tracking for the
            plugin's configuration.
        """
        return None

    # noinspection PyMethodMayBeStatic
    def on_settings_migrate(self, target, current):
        """
        Called by OctoPrint if it detects that the installed version of the plugin necessitates a higher settings version
        than the one currently stored in _config.yaml. Will also be called if the settings data stored in config.yaml
        doesn't have version information, in which case the ``current`` parameter will be None.

        Your plugin's implementation should take care of migrating any data by utilizing self._settings. OctoPrint
        will take care of saving any changes to disk by calling `self._settings.save()` after returning from this method.

        This method will be called before your plugin's :func:`on_settings_initialized` method, with all injections already
        having taken place. You can therefore depend on the configuration having been migrated by the time
        :func:`on_settings_initialized` is called.

        Arguments:
            target (int): The settings format version the plugin requires, this should always be the same value as
                          returned by :func:`get_settings_version`.
            current (int or None): The settings format version as currently stored in config.yaml. May be None if
                          no version information can be found.
        """
        pass

    def on_settings_cleanup(self):
        """
        Called after migration and initialization but before call to :func:`on_settings_initialized`.

        Plugins may overwrite this method to perform additional clean up tasks.

        The default implementation just minimizes the data persisted on disk to only contain
        the differences to the defaults (in case the current data was persisted with an older
        version of OctoPrint that still duplicated default data).

        .. versionadded:: 1.3.0
        """
        import octoprint.util
        from octoprint.settings import NoSuchSettingsPath

        try:
            # let's fetch the current persisted config (so only the data on disk,
            # without the defaults)
            config = self._settings.get_all_data(
                merged=False, incl_defaults=False, error_on_path=True
            )
        except NoSuchSettingsPath:
            # no config persisted, nothing to do => get out of here
            return

        if config is None:
            # config is set to None, that doesn't make sense, kill it and leave
            self._settings.clean_all_data()
            return

        if self.config_version_key in config and config[self.config_version_key] is None:
            # delete None entries for config version - it's the default, no need
            del config[self.config_version_key]

        # calculate a minimal diff between the settings and the current config -
        # anything already in the settings will be removed from the persisted
        # config, no need to duplicate it
        defaults = self.get_settings_defaults()
        diff = octoprint.util.dict_minimal_mergediff(defaults, config)

        if not diff:
            # no diff to defaults, no need to have anything persisted
            self._settings.clean_all_data()
        else:
            # diff => persist only that
            self._settings.set([], diff)

    def on_settings_initialized(self):
        """
        Called after the settings have been initialized and - if necessary - also been migrated through a call to
        func:`on_settings_migrate`.

        This method will always be called after the `initialize` method.
        """
        pass


class EventHandlerPlugin(OctoPrintPlugin):
    """
    The ``EventHandlerPlugin`` mixin allows OctoPrint plugins to react to any of :ref:`OctoPrint's events <sec-events>`.
    OctoPrint will call the :func:`on_event` method for any event fired on its internal event bus, supplying the
    event type and the associated payload. Please note that until your plugin returns from that method, further event
    processing within OctoPrint will block - the event queue itself is run asynchronously from the rest of OctoPrint,
    but the processing of the events within the queue itself happens consecutively.

    This mixin is especially interesting for plugins which want to react on things like print jobs finishing, timelapse
    videos rendering etc.
    """

    # noinspection PyMethodMayBeStatic
    def on_event(self, event, payload):
        """
        Called by OctoPrint upon processing of a fired event on the platform.

        .. warning::

           Do not perform long-running or even blocking operations in your implementation or you **will** block and break the server.

        Arguments:
            event (str): The type of event that got fired, see :ref:`the list of events <sec-events-available_events>`
                for possible values
            payload (dict): The payload as provided with the event
        """
        pass


class SlicerPlugin(OctoPrintPlugin):
    """
    Via the ``SlicerPlugin`` mixin plugins can add support for slicing engines to be used by OctoPrint.

    """

    # noinspection PyMethodMayBeStatic
    def is_slicer_configured(self):
        """
        Unless the return value of this method is ``True``, OctoPrint will not register the slicer within the slicing
        sub system upon startup. Plugins may use this to do some start up checks to verify that e.g. the path to
        a slicing binary as set and the binary is executable, or credentials of a cloud slicing platform are properly
        entered etc.
        """
        return False

    # noinspection PyMethodMayBeStatic
    def get_slicer_properties(self):
        """
        Plugins should override this method to return a ``dict`` containing a bunch of meta data about the implemented slicer.

        The expected keys in the returned ``dict`` have the following meaning:

        type
            The type identifier to use for the slicer. This should be a short unique lower case string which will be
            used to store slicer profiles under or refer to the slicer programmatically or from the API.
        name
            The human readable name of the slicer. This will be displayed to the user during slicer selection.
        same_device
            True if the slicer runs on the same device as OctoPrint, False otherwise. Slicers running on the same
            device will not be allowed to slice on systems with less than two CPU cores (or an unknown number) while a
            print is running due to performance reasons. Slice requests against slicers running on the same device and
            less than two cores will result in an error.
        progress_report
            ``True`` if the slicer can report back slicing progress to OctoPrint ``False`` otherwise.
        source_file_types
            A list of file types this slicer supports as valid origin file types. These are file types as found in the
            paths within the extension tree. Plugins may add additional file types through the :ref:`sec-plugins-hook-filemanager-extensiontree` hook.
            The system will test source files contains in incoming slicing requests via :meth:`octoprint.filemanager.valid_file_type` against the
            targeted slicer's ``source_file_types``.
        destination_extension
            The possible extensions of slicing result files.

        Returns:
            dict: A dict describing the slicer as outlined above.
        """
        return {
            "type": None,
            "name": None,
            "same_device": True,
            "progress_report": False,
            "source_file_types": ["model"],
            "destination_extensions": ["gco", "gcode", "g"],
        }

    # noinspection PyMethodMayBeStatic
    def get_slicer_extension_tree(self):
        """
        Fetch additional entries to put into the extension tree for accepted files

        By default, a subtree for ``model`` files with ``stl`` extension is returned. Slicers who want to support
        additional/other file types will want to override this.

        For the extension tree format, take a look at the docs of the :ref:`octoprint.filemanager.extension_tree hook <sec-plugins-hook-filemanager-extensiontree>`.

        Returns: (dict) a dictionary containing a valid extension subtree.

        .. versionadded:: 1.3.11
        """
        from octoprint.filemanager import ContentTypeMapping

        return {"model": {"stl": ContentTypeMapping(["stl"], "application/sla")}}

    def get_slicer_profiles(self, profile_path):
        """
        Fetch all :class:`~octoprint.slicing.SlicingProfile` stored for this slicer.

        For compatibility reasons with existing slicing plugins this method defaults to returning profiles parsed from
        .profile files in the plugin's ``profile_path``, utilizing the :func:`SlicingPlugin.get_slicer_profile` method
        of the plugin implementation.

        Arguments:
            profile_path (str): The base folder where OctoPrint stores this slicer plugin's profiles

        .. versionadded:: 1.3.7
        """

        from os import scandir

        import octoprint.util

        profiles = {}
        for entry in scandir(profile_path):
            if not entry.name.endswith(".profile") or octoprint.util.is_hidden_path(
                entry.name
            ):
                # we are only interested in profiles and no hidden files
                continue

            profile_name = entry.name[: -len(".profile")]
            profiles[profile_name] = self.get_slicer_profile(entry.path)
        return profiles

    # noinspection PyMethodMayBeStatic
    def get_slicer_profiles_lastmodified(self, profile_path):
        """
        .. versionadded:: 1.3.0
        """
        import os

        lms = [os.stat(profile_path).st_mtime]
        lms += [
            os.stat(entry.path).st_mtime
            for entry in os.scandir(profile_path)
            if entry.name.endswith(".profile")
        ]
        return max(lms)

    # noinspection PyMethodMayBeStatic
    def get_slicer_default_profile(self):
        """
        Should return a :class:`~octoprint.slicing.SlicingProfile` containing the default slicing profile to use with
        this slicer if no other profile has been selected.

        Returns:
            SlicingProfile: The :class:`~octoprint.slicing.SlicingProfile` containing the default slicing profile for
                this slicer.
        """
        return None

    # noinspection PyMethodMayBeStatic
    def get_slicer_profile(self, path):
        """
        Should return a :class:`~octoprint.slicing.SlicingProfile` parsed from the slicing profile stored at the
        indicated ``path``.

        Arguments:
            path (str): The absolute path from which to read the slicing profile.

        Returns:
            SlicingProfile: The specified slicing profile.
        """
        return None

    # noinspection PyMethodMayBeStatic
    def save_slicer_profile(self, path, profile, allow_overwrite=True, overrides=None):
        """
        Should save the provided :class:`~octoprint.slicing.SlicingProfile` to the indicated ``path``, after applying
        any supplied ``overrides``. If a profile is already saved under the indicated path and ``allow_overwrite`` is
        set to False (defaults to True), an :class:`IOError` should be raised.

        Arguments:
            path (str): The absolute path to which to save the profile.
            profile (SlicingProfile): The profile to save.
            allow_overwrite (boolean): Whether to allow to overwrite an existing profile at the indicated path (True,
                default) or not (False). If a profile already exists on the path and this is False an
                :class:`IOError` should be raised.
            overrides (dict): Profile overrides to apply to the ``profile`` before saving it
        """
        pass

    # noinspection PyMethodMayBeStatic
    def do_slice(
        self,
        model_path,
        printer_profile,
        machinecode_path=None,
        profile_path=None,
        position=None,
        on_progress=None,
        on_progress_args=None,
        on_progress_kwargs=None,
    ):
        """
        Called by OctoPrint to slice ``model_path`` for the indicated ``printer_profile``. If the ``machinecode_path`` is ``None``,
        slicer implementations should generate it from the provided ``model_path``.

        If provided, the ``profile_path`` is guaranteed by OctoPrint to be a serialized slicing profile created through the slicing
        plugin's own :func:`save_slicer_profile` method.

        If provided, ``position`` will be a ``dict`` containing and ``x`` and a ``y`` key, indicating the position
        the center of the model on the print bed should have in the final sliced machine code. If not provided, slicer
        implementations should place the model in the center of the print bed.

        ``on_progress`` will be a callback which expects an additional keyword argument ``_progress`` with the current
        slicing progress which - if progress reporting is supported - the slicing plugin should call like the following:

        .. code-block:: python

           if on_progress is not None:
               if on_progress_args is None:
                   on_progress_args = ()
               if on_progress_kwargs is None:
                   on_progress_kwargs = dict()

               on_progress_kwargs["_progress"] = your_plugins_slicing_progress
               on_progress(*on_progress_args, **on_progress_kwargs)

        Please note that both ``on_progress_args`` and ``on_progress_kwargs`` as supplied by OctoPrint might be ``None``,
        so always make sure to initialize those values to sane defaults like depicted above before invoking the callback.

        In order to support external cancellation of an ongoing slicing job via :func:`cancel_slicing`, implementations
        should make sure to track the started jobs via the ``machinecode_path``, if provided.

        The method should return a 2-tuple consisting of a boolean ``flag`` indicating whether the slicing job was
        finished successfully (True) or not (False) and a ``result`` depending on the success of the slicing job.

        For jobs that finished successfully, ``result`` should be a :class:`dict` containing additional information
        about the slicing job under the following keys:

        analysis
            Analysis result of the generated machine code as returned by the slicer itself. This should match the
            data structure described for the analysis queue of the matching machine code format, e.g.
            :class:`~octoprint.filemanager.analysis.GcodeAnalysisQueue` for GCODE files.

        For jobs that did not finish successfully (but not due to being cancelled!), ``result`` should be a :class:`str`
        containing a human readable reason for the error.

        If the job gets cancelled, a :class:`~octoprint.slicing.SlicingCancelled` exception should be raised.

        Returns:
            tuple: A 2-tuple (boolean, object) as outlined above.

        Raises:
            SlicingCancelled: The slicing job was cancelled (via :meth:`cancel_slicing`).
        """
        pass

    # noinspection PyMethodMayBeStatic
    def cancel_slicing(self, machinecode_path):
        """
        Cancels the slicing to the indicated file.

        Arguments:
            machinecode_path (str): The absolute path to the machine code file to which to stop slicing to.
        """
        pass


class ProgressPlugin(OctoPrintPlugin):
    """
    Via the ``ProgressPlugin`` mixing plugins can let themselves be called upon progress in print jobs or slicing jobs,
    limited to minimally 1% steps.
    """

    # noinspection PyMethodMayBeStatic
    def on_print_progress(self, storage, path, progress):
        """
        Called by OctoPrint on minimally 1% increments during a running print job.

        :param string storage:  Location of the file
        :param string path:     Path of the file
        :param int progress:    Current progress as a value between 0 and 100
        """
        pass

    # noinspection PyMethodMayBeStatic
    def on_slicing_progress(
        self,
        slicer,
        source_location,
        source_path,
        destination_location,
        destination_path,
        progress,
    ):
        """
        Called by OctoPrint on minimally 1% increments during a running slicing job.

        :param string slicer:               Key of the slicer reporting the progress
        :param string source_location:      Location of the source file
        :param string source_path:          Path of the source file
        :param string destination_location: Location the destination file
        :param string destination_path:     Path of the destination file
        :param int progress:                Current progress as a value between 0 and 100
        """
        pass



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import errno
import logging
import subprocess
import sys

from ..exceptions import ConfigurationInvalid


def _get_git_executables():
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
    return GITS


def _git(args, cwd, hide_stderr=False):
    commands = _get_git_executables()

    for c in commands:
        try:
            p = subprocess.Popen(
                [c] + args,
                cwd=cwd,
                stdout=subprocess.PIPE,
                stderr=(subprocess.PIPE if hide_stderr else None),
            )
            break
        except OSError:
            e = sys.exc_info()[1]
            if e.errno == errno.ENOENT:
                continue
            return None, None
    else:
        return None, None

    stdout = p.communicate()[0].strip()
    if sys.version >= "3":
        stdout = stdout.decode()

    if p.returncode != 0:
        return p.returncode, None

    return p.returncode, stdout


def get_latest(target, check, online=True, *args, **kwargs):
    checkout_folder = check.get("checkout_folder")
    if checkout_folder is None:
        raise ConfigurationInvalid(
            "Update configuration for {} of type git_commit needs checkout_folder set and not None".format(
                target
            )
        )

    returncode, local_commit = _git(["rev-parse", "@{0}"], checkout_folder)
    if returncode != 0:
        return None, True

    information = {
        "local": {"name": "Commit %s" % local_commit, "value": local_commit},
        "remote": {"name": "?", "value": "?"},
        "needs_online": not check.get("offline", False),
    }
    if not online and information["needs_online"]:
        return information, True

    returncode, _ = _git(["fetch"], checkout_folder)
    if returncode != 0:
        return information, True

    returncode, remote_commit = _git(["rev-parse", "@{u}"], checkout_folder)
    if returncode != 0:
        return information, True

    returncode, base = _git(["merge-base", "@{0}", "@{u}"], checkout_folder)
    if returncode != 0:
        return information, True

    if local_commit == remote_commit or remote_commit == base:
        information["remote"] = {
            "name": "Commit %s" % local_commit,
            "value": local_commit,
        }
        is_current = True
    else:
        information["remote"] = {
            "name": "Commit %s" % remote_commit,
            "value": remote_commit,
        }
        is_current = local_commit == remote_commit

    logger = logging.getLogger(
        "octoprint.plugins.softwareupdate.version_checks.git_commit"
    )
    logger.debug(f"Target: {target}, local: {local_commit}, remote: {remote_commit}")

    return information, is_current



__author__ = "Gina Häußge <osd@foosel.net> based on work by David Braam"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2013 David Braam, Gina Häußge - Released under terms of the AGPLv3 License"


import base64
import codecs
import io
import logging
import math
import os
import re
import zlib


class Vector3D:
    """
    3D vector value

    Supports addition, subtraction and multiplication with a scalar value (float, int) as well as calculating the
    length of the vector.

    Examples:

    >>> a = Vector3D(1.0, 1.0, 1.0)
    >>> b = Vector3D(4.0, 4.0, 4.0)
    >>> a + b == Vector3D(5.0, 5.0, 5.0)
    True
    >>> b - a == Vector3D(3.0, 3.0, 3.0)
    True
    >>> abs(a - b) == Vector3D(3.0, 3.0, 3.0)
    True
    >>> a * 2 == Vector3D(2.0, 2.0, 2.0)
    True
    >>> a * 2 == 2 * a
    True
    >>> a.length == math.sqrt(a.x ** 2 + a.y ** 2 + a.z ** 2)
    True
    >>> copied_a = Vector3D(a)
    >>> a == copied_a
    True
    >>> copied_a.x == a.x and copied_a.y == a.y and copied_a.z == a.z
    True
    """

    def __init__(self, *args):
        if len(args) == 3:
            (self.x, self.y, self.z) = args

        elif len(args) == 1:
            # copy constructor
            other = args[0]
            if not isinstance(other, Vector3D):
                raise ValueError("Object to copy must be a Vector3D instance")

            self.x = other.x
            self.y = other.y
            self.z = other.z

    @property
    def length(self):
        return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    def __add__(self, other):
        try:
            if len(other) == 3:
                return Vector3D(self.x + other[0], self.y + other[1], self.z + other[2])
        except TypeError:
            # doesn't look like a 3-tuple
            pass

        try:
            return Vector3D(self.x + other.x, self.y + other.y, self.z + other.z)
        except AttributeError:
            # also doesn't look like a Vector3D
            pass

        raise TypeError(
            "other must be a Vector3D instance or a list or tuple of length 3"
        )

    def __sub__(self, other):
        try:
            if len(other) == 3:
                return Vector3D(self.x - other[0], self.y - other[1], self.z - other[2])
        except TypeError:
            # doesn't look like a 3-tuple
            pass

        try:
            return Vector3D(self.x - other.x, self.y - other.y, self.z - other.z)
        except AttributeError:
            # also doesn't look like a Vector3D
            pass

        raise TypeError(
            "other must be a Vector3D instance or a list or tuple of length 3"
        )

    def __mul__(self, other):
        try:
            return Vector3D(self.x * other, self.y * other, self.z * other)
        except TypeError:
            # doesn't look like a scalar
            pass

        raise ValueError("other must be a float or int value")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __abs__(self):
        return Vector3D(abs(self.x), abs(self.y), abs(self.z))

    def __eq__(self, other):
        if not isinstance(other, Vector3D):
            return False
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __str__(self):
        return "Vector3D(x={}, y={}, z={}, length={})".format(
            self.x, self.y, self.z, self.length
        )


class MinMax3D:
    """
    Tracks minimum and maximum of recorded values

    Examples:

    >>> minmax = MinMax3D()
    >>> minmax.record(Vector3D(2.0, 2.0, 2.0))
    >>> minmax.min.x == 2.0 == minmax.max.x and minmax.min.y == 2.0 == minmax.max.y and minmax.min.z == 2.0 == minmax.max.z
    True
    >>> minmax.record(Vector3D(1.0, 2.0, 3.0))
    >>> minmax.min.x == 1.0 and minmax.min.y == 2.0 and minmax.min.z == 2.0
    True
    >>> minmax.max.x == 2.0 and minmax.max.y == 2.0 and minmax.max.z == 3.0
    True
    >>> minmax.size == Vector3D(1.0, 0.0, 1.0)
    True
    >>> empty = MinMax3D()
    >>> empty.size == Vector3D(0.0, 0.0, 0.0)
    True
    >>> weird = MinMax3D(min_z=-1.0)
    >>> weird.record(Vector3D(2.0, 2.0, 2.0))
    >>> weird.record(Vector3D(1.0, 2.0, 3.0))
    >>> weird.min.z == -1.0
    True
    >>> weird.size == Vector3D(1.0, 0.0, 4.0)
    True
    """

    def __init__(
        self,
        min_x=None,
        min_y=None,
        min_z=None,
        max_x=None,
        max_y=None,
        max_z=None,
    ):
        min_x = min_x if min_x is not None else float("inf")
        min_y = min_y if min_y is not None else float("inf")
        min_z = min_z if min_z is not None else float("inf")
        max_x = max_x if max_x is not None else -float("inf")
        max_y = max_y if max_y is not None else -float("inf")
        max_z = max_z if max_z is not None else -float("inf")

        self.min = Vector3D(min_x, min_y, min_z)
        self.max = Vector3D(max_x, max_y, max_z)

    def record(self, coordinate):
        """
        Records the coordinate, storing the min and max values.

        The input vector components must not be None.
        """
        self.min.x = min(self.min.x, coordinate.x)
        self.min.y = min(self.min.y, coordinate.y)
        self.min.z = min(self.min.z, coordinate.z)
        self.max.x = max(self.max.x, coordinate.x)
        self.max.y = max(self.max.y, coordinate.y)
        self.max.z = max(self.max.z, coordinate.z)

    @property
    def size(self):
        result = Vector3D()
        for c in "xyz":
            min = getattr(self.min, c)
            max = getattr(self.max, c)
            value = abs(max - min) if max >= min else 0.0
            setattr(result, c, value)
        return result


class AnalysisAborted(Exception):
    def __init__(self, reenqueue=True, *args, **kwargs):
        self.reenqueue = reenqueue
        Exception.__init__(self, *args, **kwargs)


regex_command = re.compile(
    r"^\s*((?P<codeGM>[GM]\d+)(\.(?P<subcode>\d+))?|(?P<codeT>T)(?P<tool>\d+))"
)
"""Regex for a GCODE command."""


class gcode:
    def __init__(self, incl_layers=False, progress_callback=None):
        self._logger = logging.getLogger(__name__)
        self.extrusionAmount = [0]
        self.extrusionVolume = [0]
        self.totalMoveTimeMinute = 0
        self.filename = None
        self._abort = False
        self._reenqueue = True
        self._filamentDiameter = 0
        self._minMax = MinMax3D()
        self._progress_callback = progress_callback

        self._incl_layers = incl_layers
        self._layers = []
        self._current_layer = None

    def _track_layer(self, pos, arc=None):
        if not self._incl_layers:
            return

        if self._current_layer is None or self._current_layer["z"] != pos.z:
            self._current_layer = {"z": pos.z, "minmax": MinMax3D(), "commands": 1}
            self._layers.append(self._current_layer)

        elif self._current_layer:
            self._current_layer["minmax"].record(pos)
            if arc is not None:
                self._addArcMinMax(
                    self._current_layer["minmax"],
                    arc["startAngle"],
                    arc["endAngle"],
                    arc["center"],
                    arc["radius"],
                )

    def _track_command(self):
        if self._current_layer:
            self._current_layer["commands"] += 1

    @property
    def dimensions(self):
        size = self._minMax.size
        return {"width": size.x, "depth": size.y, "height": size.z}

    @property
    def printing_area(self):
        return {
            "minX": None if math.isinf(self._minMax.min.x) else self._minMax.min.x,
            "minY": None if math.isinf(self._minMax.min.y) else self._minMax.min.y,
            "minZ": None if math.isinf(self._minMax.min.z) else self._minMax.min.z,
            "maxX": None if math.isinf(self._minMax.max.x) else self._minMax.max.x,
            "maxY": None if math.isinf(self._minMax.max.y) else self._minMax.max.y,
            "maxZ": None if math.isinf(self._minMax.max.z) else self._minMax.max.z,
        }

    @property
    def layers(self):
        return [
            {
                "num": num + 1,
                "z": layer["z"],
                "commands": layer["commands"],
                "bounds": {
                    "minX": layer["minmax"].min.x,
                    "maxX": layer["minmax"].max.x,
                    "minY": layer["minmax"].min.y,
                    "maxY": layer["minmax"].max.y,
                },
            }
            for num, layer in enumerate(self._layers)
        ]

    def load(
        self,
        filename,
        throttle=None,
        speedx=6000,
        speedy=6000,
        offsets=None,
        max_extruders=10,
        g90_extruder=False,
        bed_z=0.0,
    ):
        self._minMax.min.z = bed_z
        if os.path.isfile(filename):
            self.filename = filename
            self._fileSize = os.stat(filename).st_size

            with codecs.open(filename, encoding="utf-8", errors="replace") as f:
                self._load(
                    f,
                    throttle=throttle,
                    speedx=speedx,
                    speedy=speedy,
                    offsets=offsets,
                    max_extruders=max_extruders,
                    g90_extruder=g90_extruder,
                )

    def abort(self, reenqueue=True):
        self._abort = True
        self._reenqueue = reenqueue

    def _load(
        self,
        gcodeFile,
        throttle=None,
        speedx=6000,
        speedy=6000,
        offsets=None,
        max_extruders=10,
        g90_extruder=False,
    ):
        lineNo = 0
        readBytes = 0
        pos = Vector3D(0.0, 0.0, 0.0)
        currentE = [0.0]
        totalExtrusion = [0.0]
        maxExtrusion = [0.0]
        currentExtruder = 0
        totalMoveTimeMinute = 0.0
        relativeE = False
        relativeMode = False
        duplicationMode = False
        scale = 1.0
        fwretractTime = 0
        fwretractDist = 0
        fwrecoverTime = 0
        feedrate = min(speedx, speedy)
        if feedrate == 0:
            # some somewhat sane default if axes speeds are insane...
            feedrate = 2000

        if offsets is None or not isinstance(offsets, (list, tuple)):
            offsets = []
        if len(offsets) < max_extruders:
            offsets += [(0, 0)] * (max_extruders - len(offsets))

        for line in gcodeFile:
            if self._abort:
                raise AnalysisAborted(reenqueue=self._reenqueue)
            lineNo += 1
            readBytes += len(line.encode("utf-8"))

            if isinstance(gcodeFile, (io.IOBase, codecs.StreamReaderWriter)):
                percentage = readBytes / self._fileSize
            elif isinstance(gcodeFile, (list)):
                percentage = lineNo / len(gcodeFile)
            else:
                percentage = None

            try:
                if (
                    self._progress_callback is not None
                    and (lineNo % 1000 == 0)
                    and percentage is not None
                ):
                    self._progress_callback(percentage)
            except Exception as exc:
                self._logger.debug(
                    "Progress callback %r error: %s", self._progress_callback, exc
                )

            if ";" in line:
                comment = line[line.find(";") + 1 :].strip()
                if comment.startswith("filament_diameter"):
                    # Slic3r
                    filamentValue = comment.split("=", 1)[1].strip()
                    try:
                        self._filamentDiameter = float(filamentValue)
                    except ValueError:
                        try:
                            self._filamentDiameter = float(
                                filamentValue.split(",")[0].strip()
                            )
                        except ValueError:
                            self._filamentDiameter = 0.0
                elif comment.startswith("CURA_PROFILE_STRING") or comment.startswith(
                    "CURA_OCTO_PROFILE_STRING"
                ):
                    # Cura 15.04.* & OctoPrint Cura plugin
                    if comment.startswith("CURA_PROFILE_STRING"):
                        prefix = "CURA_PROFILE_STRING:"
                    else:
                        prefix = "CURA_OCTO_PROFILE_STRING:"

                    curaOptions = self._parseCuraProfileString(comment, prefix)
                    if "filament_diameter" in curaOptions:
                        try:
                            self._filamentDiameter = float(
                                curaOptions["filament_diameter"]
                            )
                        except ValueError:
                            self._filamentDiameter = 0.0
                elif comment.startswith("filamentDiameter,"):
                    # Simplify3D
                    filamentValue = comment.split(",", 1)[1].strip()
                    try:
                        self._filamentDiameter = float(filamentValue)
                    except ValueError:
                        self._filamentDiameter = 0.0
                line = line[0 : line.find(";")]

            match = regex_command.search(line)
            gcode = tool = None
            if match:
                values = match.groupdict()
                if "codeGM" in values and values["codeGM"]:
                    gcode = values["codeGM"]
                elif "codeT" in values and values["codeT"]:
                    gcode = values["codeT"]
                    tool = int(values["tool"])

            # G codes
            if gcode in ("G0", "G1", "G00", "G01"):  # Move
                x = getCodeFloat(line, "X")
                y = getCodeFloat(line, "Y")
                z = getCodeFloat(line, "Z")
                e = getCodeFloat(line, "E")
                f = getCodeFloat(line, "F")

                if x is not None or y is not None or z is not None:
                    # this is a move
                    move = True
                else:
                    # print head stays on position
                    move = False

                oldPos = pos

                # Use new coordinates if provided. If not provided, use prior coordinates (minus tool offset)
                # in absolute and 0.0 in relative mode.
                newPos = Vector3D(
                    x * scale if x is not None else (0.0 if relativeMode else pos.x),
                    y * scale if y is not None else (0.0 if relativeMode else pos.y),
                    z * scale if z is not None else (0.0 if relativeMode else pos.z),
                )

                if relativeMode:
                    # Relative mode: add to current position
                    pos += newPos
                else:
                    # Absolute mode: apply tool offsets
                    pos = newPos

                if f is not None and f != 0:
                    feedrate = f

                if e is not None:
                    if relativeMode or relativeE:
                        # e is already relative, nothing to do
                        pass
                    else:
                        e -= currentE[currentExtruder]

                    # If move with extrusion, calculate new min/max coordinates of model
                    if e > 0 and move:
                        # extrusion and move -> oldPos & pos relevant for print area & dimensions
                        self._minMax.record(oldPos)
                        self._minMax.record(pos)

                    totalExtrusion[currentExtruder] += e
                    currentE[currentExtruder] += e
                    maxExtrusion[currentExtruder] = max(
                        maxExtrusion[currentExtruder], totalExtrusion[currentExtruder]
                    )

                    if currentExtruder == 0 and len(currentE) > 1 and duplicationMode:
                        # Copy first extruder length to other extruders
                        for i in range(1, len(currentE)):
                            totalExtrusion[i] += e
                            currentE[i] += e
                            maxExtrusion[i] = max(maxExtrusion[i], totalExtrusion[i])
                else:
                    e = 0

                # move time in x, y, z, will be 0 if no movement happened
                moveTimeXYZ = abs((oldPos - pos).length / feedrate)

                # time needed for extruding, will be 0 if no extrusion happened
                extrudeTime = abs(e / feedrate)

                # time to add is maximum of both
                totalMoveTimeMinute += max(moveTimeXYZ, extrudeTime)

                # process layers if there's extrusion
                if e:
                    self._track_layer(pos)

            if gcode in ("G2", "G3", "G02", "G03"):  # Arc Move
                x = getCodeFloat(line, "X")
                y = getCodeFloat(line, "Y")
                z = getCodeFloat(line, "Z")
                e = getCodeFloat(line, "E")
                i = getCodeFloat(line, "I")
                j = getCodeFloat(line, "J")
                r = getCodeFloat(line, "R")
                f = getCodeFloat(line, "F")

                # this is a move or print head stays on position
                move = (
                    x is not None
                    or y is not None
                    or z is not None
                    or i is not None
                    or j is not None
                    or r is not None
                )

                oldPos = pos

                # Use new coordinates if provided. If not provided, use prior coordinates (minus tool offset)
                # in absolute and 0.0 in relative mode.
                newPos = Vector3D(
                    x * scale if x is not None else (0.0 if relativeMode else pos.x),
                    y * scale if y is not None else (0.0 if relativeMode else pos.y),
                    z * scale if z is not None else (0.0 if relativeMode else pos.z),
                )

                if relativeMode:
                    # Relative mode: add to current position
                    pos += newPos
                else:
                    # Absolute mode: apply tool offsets
                    pos = newPos

                if f is not None and f != 0:
                    feedrate = f

                # get radius and offset
                i = 0 if i is None else i
                j = 0 if j is None else j
                r = math.sqrt(i * i + j * j) if r is None else r

                # calculate angles
                centerArc = Vector3D(oldPos.x + i, oldPos.y + j, oldPos.z)
                startAngle = math.atan2(oldPos.y - centerArc.y, oldPos.x - centerArc.x)
                endAngle = math.atan2(pos.y - centerArc.y, pos.x - centerArc.x)

                if gcode == "G2":
                    startAngle, endAngle = endAngle, startAngle
                if startAngle < 0:
                    startAngle += math.pi * 2
                if endAngle < 0:
                    endAngle += math.pi * 2

                # from now on we only think in counter-clockwise direction

                if e is not None:
                    if relativeMode or relativeE:
                        # e is already relative, nothing to do
                        pass
                    else:
                        e -= currentE[currentExtruder]

                    # If move with extrusion, calculate new min/max coordinates of model
                    if e > 0 and move:
                        # extrusion and move -> oldPos & pos relevant for print area & dimensions
                        self._minMax.record(oldPos)
                        self._minMax.record(pos)
                        self._addArcMinMax(
                            self._minMax, startAngle, endAngle, centerArc, r
                        )

                    totalExtrusion[currentExtruder] += e
                    currentE[currentExtruder] += e
                    maxExtrusion[currentExtruder] = max(
                        maxExtrusion[currentExtruder], totalExtrusion[currentExtruder]
                    )

                    if currentExtruder == 0 and len(currentE) > 1 and duplicationMode:
                        # Copy first extruder length to other extruders
                        for i in range(1, len(currentE)):
                            totalExtrusion[i] += e
                            currentE[i] += e
                            maxExtrusion[i] = max(maxExtrusion[i], totalExtrusion[i])
                else:
                    e = 0

                # move time in x, y, z, will be 0 if no movement happened
                moveTimeXYZ = abs((oldPos - pos).length / feedrate)

                # time needed for extruding, will be 0 if no extrusion happened
                extrudeTime = abs(e / feedrate)

                # time to add is maximum of both
                totalMoveTimeMinute += max(moveTimeXYZ, extrudeTime)

                # process layers if there's extrusion
                if e:
                    self._track_layer(
                        pos,
                        {
                            "startAngle": startAngle,
                            "endAngle": endAngle,
                            "center": centerArc,
                            "radius": r,
                        },
                    )

            elif gcode == "G4":  # Delay
                S = getCodeFloat(line, "S")
                if S is not None:
                    totalMoveTimeMinute += S / 60
                P = getCodeFloat(line, "P")
                if P is not None:
                    totalMoveTimeMinute += P / 60 / 1000
            elif gcode == "G10":  # Firmware retract
                totalMoveTimeMinute += fwretractTime
            elif gcode == "G11":  # Firmware retract recover
                totalMoveTimeMinute += fwrecoverTime
            elif gcode == "G20":  # Units are inches
                scale = 25.4
            elif gcode == "G21":  # Units are mm
                scale = 1.0
            elif gcode == "G28":  # Home
                x = getCodeFloat(line, "X")
                y = getCodeFloat(line, "Y")
                z = getCodeFloat(line, "Z")
                origin = Vector3D(0.0, 0.0, 0.0)
                if x is None and y is None and z is None:
                    pos = origin
                else:
                    pos = Vector3D(pos)
                    if x is not None:
                        pos.x = origin.x
                    if y is not None:
                        pos.y = origin.y
                    if z is not None:
                        pos.z = origin.z
            elif gcode == "G90":  # Absolute position
                relativeMode = False
                if g90_extruder:
                    relativeE = False
            elif gcode == "G91":  # Relative position
                relativeMode = True
                if g90_extruder:
                    relativeE = True
            elif gcode == "G92":
                x = getCodeFloat(line, "X")
                y = getCodeFloat(line, "Y")
                z = getCodeFloat(line, "Z")
                e = getCodeFloat(line, "E")

                if e is None and x is None and y is None and z is None:
                    # no parameters, set all axis to 0
                    currentE[currentExtruder] = 0.0
                    pos.x = 0.0
                    pos.y = 0.0
                    pos.z = 0.0
                else:
                    # some parameters set, only set provided axes
                    if e is not None:
                        currentE[currentExtruder] = e
                    if x is not None:
                        pos.x = x
                    if y is not None:
                        pos.y = y
                    if z is not None:
                        pos.z = z
            # M codes
            elif gcode == "M82":  # Absolute E
                relativeE = False
            elif gcode == "M83":  # Relative E
                relativeE = True
            elif gcode in ("M207", "M208"):  # Firmware retract settings
                s = getCodeFloat(line, "S")
                f = getCodeFloat(line, "F")
                if s is not None and f is not None:
                    if gcode == "M207":
                        fwretractTime = s / f
                        fwretractDist = s
                    else:
                        fwrecoverTime = (fwretractDist + s) / f
            elif gcode == "M605":  # Duplication/Mirroring mode
                s = getCodeInt(line, "S")
                if s in [2, 4, 5, 6]:
                    # Duplication / Mirroring mode selected. Printer firmware copies extrusion commands
                    # from first extruder to all other extruders
                    duplicationMode = True
                else:
                    duplicationMode = False

            # T codes
            elif tool is not None:
                if tool > max_extruders:
                    self._logger.warning(
                        "GCODE tried to select tool %d, that looks wrong, ignoring for GCODE analysis"
                        % tool
                    )
                elif tool == currentExtruder:
                    pass
                else:
                    pos.x -= (
                        offsets[currentExtruder][0]
                        if currentExtruder < len(offsets)
                        else 0
                    )
                    pos.y -= (
                        offsets[currentExtruder][1]
                        if currentExtruder < len(offsets)
                        else 0
                    )

                    currentExtruder = tool

                    pos.x += (
                        offsets[currentExtruder][0]
                        if currentExtruder < len(offsets)
                        else 0
                    )
                    pos.y += (
                        offsets[currentExtruder][1]
                        if currentExtruder < len(offsets)
                        else 0
                    )

                    if len(currentE) <= currentExtruder:
                        for _ in range(len(currentE), currentExtruder + 1):
                            currentE.append(0.0)
                    if len(maxExtrusion) <= currentExtruder:
                        for _ in range(len(maxExtrusion), currentExtruder + 1):
                            maxExtrusion.append(0.0)
                    if len(totalExtrusion) <= currentExtruder:
                        for _ in range(len(totalExtrusion), currentExtruder + 1):
                            totalExtrusion.append(0.0)

            if gcode or tool:
                self._track_command()

            if throttle is not None:
                throttle(lineNo, readBytes)
        if self._progress_callback is not None:
            self._progress_callback(100.0)

        self.extrusionAmount = maxExtrusion
        self.extrusionVolume = [0] * len(maxExtrusion)
        for i in range(len(maxExtrusion)):
            radius = self._filamentDiameter / 2
            self.extrusionVolume[i] = (
                self.extrusionAmount[i] * (math.pi * radius * radius)
            ) / 1000
        self.totalMoveTimeMinute = totalMoveTimeMinute

    def _parseCuraProfileString(self, comment, prefix):
        return {
            key: value
            for (key, value) in map(
                lambda x: x.split(b"=", 1),
                zlib.decompress(base64.b64decode(comment[len(prefix) :])).split(b"\b"),
            )
        }

    def _intersectsAngle(self, start, end, angle):
        if end < start and angle == 0:
            # angle crosses 0 degrees
            return True
        else:
            return start <= angle <= end

    def _addArcMinMax(self, minmax, startAngle, endAngle, centerArc, radius):
        startDeg = math.degrees(startAngle)
        endDeg = math.degrees(endAngle)

        if self._intersectsAngle(startDeg, endDeg, 0):
            # arc crosses positive x
            minmax.max.x = max(minmax.max.x, centerArc.x + radius)
        if self._intersectsAngle(startDeg, endDeg, 90):
            # arc crosses positive y
            minmax.max.y = max(minmax.max.y, centerArc.y + radius)
        if self._intersectsAngle(startDeg, endDeg, 180):
            # arc crosses negative x
            minmax.min.x = min(minmax.min.x, centerArc.x - radius)
        if self._intersectsAngle(startDeg, endDeg, 270):
            # arc crosses negative y
            minmax.min.y = min(minmax.min.y, centerArc.y - radius)

    def get_result(self):
        result = {
            "total_time": self.totalMoveTimeMinute,
            "extrusion_length": self.extrusionAmount,
            "extrusion_volume": self.extrusionVolume,
            "dimensions": self.dimensions,
            "printing_area": self.printing_area,
        }
        if self._incl_layers:
            result["layers"] = self.layers

        return result


def getCodeInt(line, code):
    return getCode(line, code, int)


def getCodeFloat(line, code):
    return getCode(line, code, float)


def getCode(line, code, c):
    n = line.find(code) + 1
    if n < 1:
        return None
    m = line.find(" ", n)
    try:
        if m < 0:
            result = c(line[n:])
        else:
            result = c(line[n:m])
    except ValueError:
        return None

    if math.isnan(result) or math.isinf(result):
        return None

    return result



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from octoprint.util import deprecated

from . import serializing  # noqa: F401
from .encoding import JsonEncoding, dumps, loads  # noqa: F401

dump = deprecated(
    "dump has been renamed to dumps, please adjust your implementation",
    includedoc="dump has been renamed to dumps",
    since="1.8.0",
)(dumps)



"""
This module represents OctoPrint's system and server commands.
"""

__author__ = "Johan Verrept <johan@verrept.eu>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging

from octoprint.settings import settings
from octoprint.util.commandline import CommandlineCaller, CommandlineError

# singleton
_instance = None


def system_command_manager():
    global _instance
    if _instance is None:
        _instance = SystemCommandManager()
    return _instance


class SystemCommandManager:

    SERVER_RESTART_COMMAND = "serverRestartCommand"
    SYSTEM_RESTART_COMMAND = "systemRestartCommand"
    SYSTEM_SHUTDOWN_COMMAND = "systemShutdownCommand"

    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self._caller = CommandlineCaller()

    def execute(self, command):
        if not command:
            return False

        try:
            # we run this with shell=True since we have to trust whatever
            # our admin configured as command and since we want to allow
            # shell-alike handling here...
            p = self._caller.non_blocking_call(command, shell=True)

            if p is None:
                raise CommandlineError(None, "", "")

            if p.returncode is not None:
                stdout = p.stdout.text if p is not None and p.stdout is not None else ""
                stderr = p.stderr.text if p is not None and p.stderr is not None else ""
                raise CommandlineError(p.returncode, stdout, stderr)
        except CommandlineError:
            raise
        except Exception:
            self._logger.exception(f"Error while executing command: {command}")
            raise CommandlineError(None, "", "")

        return True

    def get_command(self, cmd):
        return settings().get(["server", "commands", cmd])

    def has_command(self, cmd):
        return self.get_command(cmd) is not None

    def get_server_restart_command(self):
        return self.get_command(self.SERVER_RESTART_COMMAND)

    def get_system_restart_command(self):
        return self.get_command(self.SYSTEM_RESTART_COMMAND)

    def get_system_shutdown_command(self):
        return self.get_command(self.SYSTEM_SHUTDOWN_COMMAND)

    def has_server_restart_command(self):
        return self.get_server_restart_command() is not None

    def has_system_restart_command(self):
        return self.get_system_restart_command() is not None

    def has_system_shutdown_command(self):
        return self.get_system_shutdown_command() is not None

    def perform_server_restart(self):
        return self.execute(self.get_server_restart_command())

    def perform_system_restart(self):
        return self.execute(self.get_system_restart_command())

    def perform_system_shutdown(self):
        return self.execute(self.get_system_shutdown_command())
